"""
ISSqlSchema module.
"""


from PyQt6 import QtCore, QtWidgets  # type: ignore[import]

from pineboolib import logging, application

from pineboolib.core.utils import utils_base
from pineboolib.application.utils import check_dependencies
from pineboolib.application.database import pnsqlquery
from pineboolib.application.parsers.parser_mtd import pnmtdparser
from pineboolib.core import settings
from pineboolib.application import qsadictmodules

from typing import Iterable, Optional, Union, List, Any, Dict, cast, TYPE_CHECKING
from pineboolib.core import decorators

from pineboolib.fllegacy import flutil

from sqlalchemy.engine import base, create_engine  # type: ignore [import] # noqa: F821
from sqlalchemy.inspection import inspect  # type: ignore [import] # noqa: F821, F401
from sqlalchemy.orm import sessionmaker  # type: ignore [import] # noqa: F821

from sqlalchemy import event, pool, text  # type: ignore [import]
import sqlalchemy  # type: ignore [import] # noqa: F821, F401
import traceback


import re

from pineboolib.interfaces import itablemetadata


if TYPE_CHECKING:
    from pineboolib.application.metadata import pntablemetadata  # noqa: F401 # pragma: no cover
    from pineboolib.application.metadata import pnfieldmetadata
    from pineboolib.interfaces import iconnection  # noqa: F401 # pragma: no cover
    from sqlalchemy.engine import (  # type: ignore [import] # noqa: F821, F401
        result,  # noqa: F401
    )  # noqa: F401 # pragma: no cover

    from pineboolib.interfaces import isession  # pragma: no cover


LOGGER = logging.get_logger(__name__)

ENGINES: Dict[str, "base.Engine"] = {}


class ISqlDriver(object):
    """ISqlDriver class."""

    version_: str
    name_: str
    alias_: str
    error_list: List[str]
    _last_error: str
    db_: "iconnection.IConnection"
    _dbname: str
    db_filename: str
    mobile_: bool
    pure_python_: bool
    default_port: int
    cursor_proxy: Dict[str, "result.Result"]  # type: ignore [name-defined]
    open_: bool
    desktop_file: bool
    _true: Union[str, bool]
    _false: Union[str, bool]
    _like_true: str
    _like_false: str
    _null: str
    _text_like: str
    _text_cascade: str
    _safe_load: Dict[str, str]
    _database_not_found_keywords: List[str]
    _parse_porc: bool
    _queqe_params: Dict[str, Any]
    _create_isolation: bool

    _sqlalchemy_name: str
    _connection: "base.Connection"
    _engine: "base.Engine"
    _session: "isession.PinebooSession"
    _extra_alternative: str
    _sp_level: int
    _use_altenative_isolation_level: bool
    _use_create_table_save_points: bool
    _can_use_preping: bool
    _tables_info: Dict[str, Dict[str, Any]]

    def __init__(self):
        """Inicialize."""
        self.version_ = ""
        self.name_ = ""
        # self._connection = None
        self.error_list = []
        self.alias_ = ""
        self._dbname = ""
        self.mobile_ = False
        self.pure_python_ = False
        self.default_port = 0
        self._parse_porc = True

        self._last_error = ""
        self._text_cascade = "CASCADE"
        self._true = "1"
        self._false = "0"
        self._like_true = "1"
        self._like_false = "0"
        self._null = "Null"
        self._text_like = "::text "
        self._safe_load = {"sqlalchemy": "sqlAlchemy"}
        self._database_not_found_keywords = ["does not exist", "no existe"]
        self._queqe_params = {}
        self._create_isolation = True
        self._extra_alternative = ""
        self._connection = None  # type: ignore [assignment] # noqa: F821
        self._sqlalchemy_name = ""
        self._sp_level = 0
        self._use_altenative_isolation_level = False
        self._use_create_table_save_points = True
        self._can_use_preping = True
        self._tables_info = {}

    def safe_load(self, exit: bool = False) -> bool:
        """Return if the driver can loads dependencies safely."""
        return check_dependencies.check_dependencies(self._safe_load, exit)

    def close(self):
        """Close driver connection."""

        self.open_ = False
        if self._connection is not None:
            self._connection.close()
            self._connection = None  # type: ignore [assignment] # noqa: F821

    # def singleConnection(self) -> bool:
    #    """Return if driver uses a single connection."""
    #    return self._single_conn

    def loadConnectionString(self, name: str, host: str, port: int, usern: str, passw_: str) -> str:
        """Set special config."""

        return "%s://%s:%s@%s:%s/%s" % (self._sqlalchemy_name, usern, passw_, host, port, name)

    def connect(
        self, db_name: str, db_host: str, db_port: int, db_user_name: str, db_password: str
    ) -> Union["base.Connection", bool]:
        """Connect to database."""

        self.setDBName(db_name)
        self.safe_load(True)

        LOGGER.debug = LOGGER.trace  # type: ignore  # Send Debug output to Trace
        conn_ = self.getConn(db_name, db_host, db_port, db_user_name, db_password)

        if (
            conn_ is None and self.db_._name == "main_conn"
        ):  # Si no existe la conexión y soy main_conn
            if application.PROJECT._splash:
                application.PROJECT._splash.hide()
            if not application.PROJECT.DGI.localDesktop():
                return False

            _last_error = self.last_error()
            found = False
            for key in self._database_not_found_keywords:
                if key in _last_error:
                    found = True
                    break

            if found:
                ret = QtWidgets.QMessageBox.warning(
                    QtWidgets.QWidget(),
                    "Pineboo",
                    "La base de datos %s no existe.\n¿Desea crearla?" % db_name,
                    cast(
                        QtWidgets.QMessageBox.StandardButton,
                        QtWidgets.QMessageBox.StandardButton.Ok
                        | QtWidgets.QMessageBox.StandardButton.No,
                    ),
                )
                if ret == QtWidgets.QMessageBox.StandardButton.No:
                    return False
                else:
                    try:
                        tmp_conn = self.getAlternativeConn(
                            db_name, db_host, db_port, db_user_name, db_password
                        )
                        if tmp_conn is not None:
                            self.set_last_error_null()
                            try:
                                if self._create_isolation:
                                    tmp_conn.connection.set_isolation_level(0)
                                tmp_conn.execute("CREATE DATABASE %s" % db_name)
                                if self._create_isolation:
                                    tmp_conn.connection.set_isolation_level(1)
                            except Exception as error:
                                self.set_last_error(str(error), "LOGGIN")

                                tmp_conn.execute("ROLLBACK")
                                tmp_conn.close()
                                return False

                            tmp_conn.close()
                            conn_ = self.getConn(
                                db_name, db_host, db_port, db_user_name, db_password
                            )

                    except Exception as error:
                        LOGGER.warning(error)
                        QtWidgets.QMessageBox.information(
                            QtWidgets.QWidget(),
                            "Pineboo",
                            "ERROR: No se ha podido crear la Base de Datos %s" % db_name,
                            QtWidgets.QMessageBox.StandardButton.Ok,
                        )
                        LOGGER.error("ERROR: No se ha podido crear la Base de Datos %s", db_name)
                        return False
        if conn_ is not None:
            # if settings.CONFIG.value("ebcomportamiento/orm_enabled", False):
            #    self._engine = self.getEngine(db_name, db_host, db_port, db_user_name, db_password)
            self._connection = conn_
            self.open_ = True
            # self.session()
        else:
            LOGGER.error("connect: %s", self.last_error())

        return conn_ or False

    def setDBName(self, name: str) -> None:
        """Set DB Name."""

        self._dbname = name

    @decorators.not_implemented_warn
    def getAlternativeConn(
        self, name: str, host: str, port: int, usern: str, passw_: str
    ) -> Optional["base.Connection"]:
        """Return alternative connection."""

        return None  # pragma: no cover

    def getConn(
        self, name: str, host: str, port: int, usern: str, passw_: str, alternative=False
    ) -> Optional["base.Connection"]:
        """Return connection."""

        conn_ = None
        LOGGER.debug = LOGGER.trace  # type: ignore  # Send Debug output to Trace
        try:
            str_conn = self.loadConnectionString(name, host, port, usern, passw_)
            if str_conn in ENGINES.keys():
                LOGGER.debug(
                    "Reusing engine %s:%s/%s to %s connection", host, port, name, self.db_._name
                )
                self._engine = ENGINES[str_conn]
            else:
                if alternative:
                    str_conn += self._extra_alternative

                self.get_common_params()
                self._engine = create_engine(str_conn, **self._queqe_params)
                if self._use_altenative_isolation_level:
                    event.listen(self._engine, "connect", self.do_connect)
                    event.listen(self._engine, "begin", self.do_begin)
                    event.listen(self._engine, "savepoint", self.do_savepoint)
                    event.listen(self._engine, "rollback", self.do_rollback)
                    event.listen(self._engine, "commit", self.do_commit)

                ENGINES[str_conn] = self._engine

            if application.SHOW_CONNECTION_EVENTS:
                self.listen_engine()

            conn_ = self.connection()

        except Exception as error:
            self.set_last_error(str(error), "CONNECT")
        return conn_

    def loadSpecialConfig(self) -> None:
        """Set special config."""

        pass  # pragma: no cover

    def version(self) -> str:
        """Return version number."""
        return self.version_

    def driverName(self) -> str:
        """Return driver name."""
        return self.name_

    def is_open(self) -> bool:
        """Return if the connection is open."""
        if hasattr(self, "_engine"):
            conn_ = self.connection()
            return not conn_.closed
        return False

    def pure_python(self) -> bool:
        """Return if the driver is python only."""
        return self.pure_python_

    def mobile(self) -> bool:
        """Return if the driver is mobile ready."""
        return self.mobile_

    def DBName(self) -> str:
        """Return database name."""
        return self._dbname

    def engine(self) -> "base.Engine":
        """Return sqlAlchemy ORM engine."""
        return self._engine

    def session(self) -> "isession.PinebooSession":
        """Create a sqlAlchemy session."""
        while True:
            session_class = sessionmaker(
                bind=self.connection().execution_options(autocommit=True),
                autoflush=False,
                autocommit=True,
            )

            new_session = session_class()
            if new_session.connection().connection is not None:
                break
            else:
                LOGGER.warning("Conexión invalida capturada.Solicitando nueva")

        setattr(new_session, "_conn_name", self.db_._name)
        return new_session  # type: ignore [return-value]

    def connection(self) -> "base.Connection":
        """Return a cursor connection."""

        if self._connection is None or self._connection.closed:
            if getattr(self, "_engine", None):
                self._connection = self._engine.connect()
                if application.SHOW_CONNECTION_EVENTS:
                    event.listen(self._engine, "close", self.close_emited)
                    LOGGER.info("New connection created.\n%s", self._engine.pool.status())
            else:
                raise Exception("Engine is not loaded!")

        return self._connection

    def formatValueLike(self, type_: str, value: Any, upper: bool) -> str:
        """Return a string with the format value like."""

        util = flutil.FLUtil()
        res = "IS NULL"

        if type_ == "bool":
            value = str(value[0]).upper()
            if value == str(util.translate("application", "Sí")[0]).upper():
                res = "=%s" % self._like_true
            else:
                res = "=%s" % self._like_false

        elif type_ == "date":
            dateamd = util.dateDMAtoAMD(str(value))
            if dateamd is None:
                dateamd = ""
            res = self._text_like + "LIKE '%%" + dateamd + "'"

        elif type_ == "time":
            time_ = value.toTime()
            res = self._text_like + "LIKE '" + time_.toString(QtCore.Qt.DateFormat.ISODate) + "%%'"

        else:
            res = str(value)
            if upper:
                res = "%s" % res.upper()

            res = self._text_like + "LIKE '" + res + "%%'"

        return res

    def formatValue(self, type_: str, value: Any, upper: bool) -> Optional[Union[int, str, bool]]:
        """Return a string with the format value."""

        result_: Any = value

        if type_ in ("uint", "int", "double", "serial"):
            result_ = value or 0

        elif type_ in ("string", "stringlist", "timestamp"):
            if type_ == "string":
                value = utils_base.auto_qt_translate_text(value)
                if upper:
                    value = value.upper()

            result_ = "'%s'" % value or self._null

        elif type_ == "pixmap":
            result_ = "'%s'" % self.normalizeValue(value)

        elif type_ in ("bool", "unlock"):
            result_ = self._true if utils_base.text2bool(str(value)) else self._false

        elif type_ == "date":
            result_ = "'%s'" % str(flutil.FLUtil.dateDMAtoAMD(value))

        elif type_ == "time":
            result_ = "'%s'" % value if value else ""
        elif type_ == "json":
            if not value:
                result_ = {}

        return str(result_)

    def canOverPartition(self) -> bool:
        """Return can override partition option ready."""
        return True

    def canRegenTables(self) -> bool:
        """Return if can regenerate tables."""
        return application.PROJECT.db_admin_mode

    def nextSerialVal(self, table_name: str, field_name: str) -> int:
        """Return next serial value."""

        table_max = 0
        flseq_max = 0
        res_ = 0
        sql = "SELECT MAX(%s) FROM %s WHERE 1=1" % (field_name, table_name)
        cur = self.execute_query(sql)
        if cur is not None:
            value = cur.fetchone()

            if value is not None:
                table_max = value[0] or 0

        sql = "SELECT seq FROM flseqs WHERE tabla = '%s' AND campo ='%s'" % (table_name, field_name)
        cur = self.execute_query(sql)
        if cur is not None:
            value = cur.fetchone()

            if value is not None:
                flseq_max = value[0] or 0

        res_ = flseq_max or table_max
        res_ += 1

        str_qry = ""
        if flseq_max:
            if res_ > flseq_max:
                str_qry = "UPDATE flseqs SET seq=%s WHERE tabla = '%s' AND campo = '%s'" % (
                    res_,
                    table_name,
                    field_name,
                )
        else:
            str_qry = "INSERT INTO flseqs (tabla,campo,seq) VALUES('%s','%s',%s)" % (
                table_name,
                field_name,
                res_,
            )

        if str_qry:
            self.execute_query(str_qry)

        return res_

    def set_last_error_null(self) -> None:
        """Set lastError flag Null."""
        self._last_error = ""

    def set_last_error(self, text: str, command: str) -> None:
        """Set last error."""
        self._last_error = "%s (%s)" % (text, command)
        LOGGER.error(self._last_error)

    def last_error(self) -> str:
        """Return last error."""
        return self._last_error

    @decorators.not_implemented_warn
    def setType(self, type_: str, leng: int = 0) -> str:
        """Return type definition."""
        return ""  # pragma: no cover

    def existsTable(self, table_name: str) -> bool:
        """Return if exists a table specified by name."""

        if self._engine:
            self.set_last_error_null()
            table_list = self.tables("", table_name)
            if self.last_error():
                raise Exception("Error loading tables.")

            return table_name in table_list
        else:
            raise Exception("No engine or connection exists!")

    def is_valid_view(
        self, metadata: "pntablemetadata.PNTableMetaData", qry: "pnsqlquery.PNSqlQuery"
    ) -> bool:
        """Return if a view is valid."""

        valid = True
        if qry.select().find(".*") > -1:
            LOGGER.warning("the use of * is not allowed")
            valid = False
        else:
            meta_field_names = metadata.fieldNames()
            select_text = qry.select()
            for field_name in meta_field_names:
                if field_name not in select_text:
                    valid = False
                    LOGGER.warning(
                        "Field name %s not found on query select (%s)", field_name, select_text
                    )
                    break

        return valid

    def sqlCreateView(self, meta: "pntablemetadata.PNTableMetaData") -> str:
        """Return a sql create view."""

        sql = ""
        qry = pnsqlquery.PNSqlQuery(meta.name())
        if qry.select() and qry.from_():
            if self.is_valid_view(meta, qry):
                sql = "CREATE VIEW %s AS SELECT %s FROM %s" % (
                    meta.name(),
                    qry.select() or "*",
                    qry.from_(),
                )
        else:
            LOGGER.warning("sqlCreateView: %s.qry is empty or does not exists", meta.name())

        return sql

    @decorators.not_implemented_warn
    def sqlCreateTable(
        self, tmd: "pntablemetadata.PNTableMetaData", create_index: bool = True
    ) -> Optional[str]:
        """Return a create table query."""
        return ""  # pragma: no cover

    def mismatchedTable(self, table_name: str, metadata: "pntablemetadata.PNTableMetaData") -> bool:
        """Return if a table is mismatched."""

        dict_database: Dict[str, List[Any]] = self.recordInfo2(table_name)

        if metadata.isQuery():
            if application.USE_MISMATCHED_VIEWS:
                qry = pnsqlquery.PNSqlQuery(table_name)
                names = qry.select().split(",")
                if len(metadata.fieldNames()) == 0:
                    return False

                for name in names:
                    field_name = name.split(".")[1] if name.find(".") > -1 else name
                    if field_name not in dict_database.keys():
                        return True
            else:
                return False

        else:
            dict_metadata: Dict[str, List[Any]] = self.recordInfo(metadata)

            if len(dict_metadata.keys()) != len(dict_database.keys()):
                return True
            else:
                for name, meta in dict_metadata.items():
                    if (
                        name in dict_database.keys()
                    ):  # si falla una key, los campos en el metadata y database no son los mismos.
                        if self.notEqualsFields(dict_database[name], meta, False):
                            LOGGER.warning(
                                "Mismatched field %s.%s:\nMetadata : %s.\nDataBase : %s\n",
                                table_name,
                                name,
                                meta,
                                dict_database[name],
                            )
                            return True
                        else:
                            del dict_database[
                                name
                            ]  # dict_database mas peuqeño cada vez, para comparar mas rápido
                    else:
                        LOGGER.warning("Field %s not found in database", name)
                        return True

        return False

    @decorators.not_implemented_warn
    def recordInfo2(self, tablename: str) -> Dict[str, List[Any]]:
        """Return info from a database table."""
        return {}  # pragma: no cover

    def recordInfo(self, table_metadata: "itablemetadata.ITableMetaData") -> Dict[str, List[Any]]:
        """Obtain current cursor information on columns."""

        return dict(
            [
                (
                    field.name(),
                    [
                        field.name(),
                        field.type(),
                        not field.allowNull(),
                        field.length(),
                        field.partDecimal(),
                        field.defaultValue(),
                        field.isPrimaryKey(),
                    ],
                )
                for field in table_metadata.fieldList()
            ]
        )

    @decorators.not_implemented_warn
    def decodeSqlType(self, type_: str) -> str:
        """Return the specific field type."""
        return ""  # pragma: no cover

    def notEqualsFields(self, field_db: List[Any], field_meta: List[Any], is_query=False) -> bool:
        """Return if a field has changed."""
        # 0 name,
        # 1 type,
        # 2 allow_null,
        # 3 length,
        # 4 part_decimal,
        # 5 default_value,
        # 6 is_primary_key

        ret = 0
        try:
            if field_db[2] != field_meta[2] and not is_query:  # nulos
                if field_meta[1] != "serial":
                    if (
                        not field_meta[2] and field_meta[6]
                    ):  # Si en meta , nulo false y pk , dejamos pasar
                        pass
                    else:
                        ret = 1

            if not ret:
                db_type = field_db[1]
                meta_type = field_meta[1]

                if db_type == "string":
                    if meta_type == "string":
                        if field_db[3] not in [field_meta[3], 0, 255]:
                            ret = 2
                    elif meta_type not in ("time", "date"):
                        ret = 3

                elif db_type == "uint" and meta_type not in ("int", "uint", "serial"):
                    ret = 4
                elif db_type == "bool" and meta_type not in ("bool", "unlock"):
                    ret = 5
                elif db_type == "double" and meta_type != "double":
                    ret = 6
                elif db_type != "double" and meta_type == "double":
                    ret = 10
                elif db_type == "stringlist" and meta_type not in (
                    "stringlist",
                    "pixmap",
                    "string",
                ):
                    ret = 7
                elif db_type == "timestamp" and meta_type != "timestamp":
                    ret = 8
                elif db_type == "json" and meta_type != "json":
                    ret = 9

        except Exception as error:
            LOGGER.error("notEqualsFields %s %s (%s)", field_db, field_meta, str(error))

        if ret:
            LOGGER.warning(
                "Falla database: %s, metadata: %s fallo:%s" % (field_db, field_meta, ret)
            )

        return ret > 0

    @decorators.not_implemented_warn
    def tables(self, type_name: str = "", table_name: str = "") -> List[str]:
        """Return a tables list specified by type."""
        return []  # pragma: no cover

    def normalizeValue(self, text: str) -> str:
        """Return a database friendly text."""

        res = str(text).replace("'", "''")
        if self._parse_porc:
            res = res.replace("%", "%%")
        # res = res.replace(":", "\\:")
        return res

    def hasCheckColumn(self, metadata: "pntablemetadata.PNTableMetaData") -> bool:
        """Retrieve if MTD has a check column."""

        for field in metadata.fieldList():
            if field.isCheck() or field.name().endswith("_check_column"):
                return True

        return False

    @decorators.not_implemented_warn
    def constraintExists(self, name: str) -> bool:
        """Return if constraint exists specified by name."""
        return False  # pragma: no cover

    def remove_index(self, metadata: "pntablemetadata.PNTableMetaData", cursor: Any) -> bool:
        """Remove olds index."""

        return True

    def alterTable(self, new_metadata: "pntablemetadata.PNTableMetaData") -> bool:
        """Modify a table structure."""

        if not application.ALLOW_ALTER_TABLE:
            LOGGER.warning(
                "ALTER TABLE DESHABILITADO",
            )
            return False

        if self.hasCheckColumn(new_metadata):
            return False

        field_list = new_metadata.fieldList()
        if not field_list:
            return False

        util = flutil.FLUtil()
        table_name = new_metadata.name()

        old_field_names: List[str] = list(self.recordInfo2(table_name).keys())

        renamed_table = "%salteredtable%s" % (
            table_name,
            QtCore.QDateTime.currentDateTime().toString("ddhhssz"),
        )

        query = pnsqlquery.PNSqlQuery(None, "dbAux")

        session_ = query.db().session()
        query.db().transaction()

        if new_metadata.isQuery():
            if table_name in self.tables("Views"):
                query.exec_("DROP VIEW %s %s" % (table_name, self._text_cascade))
            elif table_name in self.tables("Tables"):
                query.exec_("DROP TABLE %s %s" % (table_name, self._text_cascade))

        else:
            if table_name not in self.tables("Tables"):
                LOGGER.warning(
                    "El metadata indica erroneamente, que %s es una tabla. Proceso cancelado.",
                    table_name,
                )
                session_.rollback()
                return False

            if not self.remove_index(new_metadata, query):
                session_.rollback()
                return False

            if not query.exec_("ALTER TABLE %s RENAME TO %s" % (table_name, renamed_table)):
                session_.rollback()
                return False

        if not self.db_.createTable(new_metadata):
            session_.rollback()
            return False

        if new_metadata.isQuery():
            session_.commit()
            return True

        cur = query.db().execute_query(
            "SELECT %s FROM %s WHERE 1=1" % (", ".join(old_field_names), renamed_table)
        )
        old_data_list = cur.fetchall() if cur else []
        # old_cursor = pnsqlcursor.PNSqlCursor(renamed_table, True, "dbAux")
        # old_cursor.setModeAccess(old_cursor.Browse)
        # old_cursor.select()
        util.createProgressDialog(
            util.translate("application", "Reestructurando registros para %s...")
            % new_metadata.alias(),
            len(old_data_list),
        )

        util.setLabelText(util.translate("application", "Recogiendo datos"))

        new_field_names: List[str] = new_metadata.fieldNames()

        list_records = []
        for number_progress, old_data in enumerate(old_data_list):
            new_buffer = []

            for new_idx, new_name in enumerate(new_field_names):
                new_field = new_metadata.field(new_name)
                if new_field is None:
                    LOGGER.warning(
                        "Field %s not found un metadata %s" % (new_name, new_metadata.name())
                    )
                    session_.rollback()

                    return False
                value = None
                if new_name in old_field_names:
                    value = old_data[old_field_names.index(new_name)]
                    if value is None:
                        if not new_field.allowNull():
                            value = new_field.defaultValue()
                        else:
                            continue
                elif not new_field.allowNull():
                    value = new_field.defaultValue()
                    if value is None:
                        if new_field.type() == "timestamp":
                            value = self.getTimeStamp()
                        else:
                            continue
                else:
                    continue
                new_buffer.append([new_field, value])

            list_records.append(new_buffer)

            util.setProgress(number_progress)

        util.setLabelText(util.translate("application", "Regenerando datos"))
        result_insert_multi = True
        if not self.insertMulti(table_name, list_records):
            session_.rollback()
            result_insert_multi = False
        else:
            session_.commit()

            if new_metadata.name() not in self.tables("Views"):
                query.exec_("DROP TABLE %s %s" % (renamed_table, self._text_cascade))

        util.destroyProgressDialog()

        return result_insert_multi

    def cascadeSupport(self) -> bool:
        """Return True if the driver support cascade."""
        return True

    def canDetectLocks(self) -> bool:
        """Return if can detect locks."""
        return True

    # def fix_query(self, val: str) -> str:
    #    """Fix values on SQL."""
    #    ret_ = val.replace("'true'", "1")
    #    ret_ = ret_.replace("'false'", "0")
    #    ret_ = ret_.replace("'0'", "0")
    #    ret_ = ret_.replace("'1'", "1")
    #    return ret_

    # def desktopFile(self) -> bool:
    #    """Return if use a file like database."""
    #    return self.desktop_file

    def execute_query(self, query: str) -> Optional["result.Result"]:  # type: ignore [name-defined]
        """Excecute a query and return result."""

        if not self.is_open():
            raise Exception("execute_query: Database not open %s" % self)

        self.set_last_error_null()
        session_ = self.db_.session()

        try:
            try:
                if query.find("::bytea") > -1:
                    return (  # Esto es necesario para no obtener error en la consulta con los bytearray
                        session_.connection()
                        .execution_options(autocommit=True)
                        .execute("""%s""" % query)
                    )
                else:
                    return session_.execute(text("""%s""" % query))
            except sqlalchemy.exc.DBAPIError as error:
                LOGGER.warning(
                    "Se ha producido un error DBAPI con la consulta %s. Ejecutando rollback necesario",
                    query,
                    stack_info=True,
                )
                session_.rollback()
                self.set_last_error(
                    "No se pudo ejecutar la query %s.\n%s" % (query, str(error)), query
                )

        except Exception as error:
            LOGGER.warning(
                "Se ha producido un error al ejecutar la consulta %s.\n%s\nSTACK APP: %s\nSTACK ERROR: %s",
                query,
                str(error),
                "".join(traceback.format_exc(limit=None)),
                "".join(traceback.format_stack(limit=None)),
            )
            self.set_last_error("No se pudo ejecutar la query %s.\n%s" % (query, str(error)), query)

        return None

    def getTimeStamp(self) -> str:
        """Return TimeStamp."""

        sql = "SELECT CURRENT_TIMESTAMP"

        cur = self.execute_query(sql)

        time_stamp_: Any = cur.fetchone() if cur else None

        if time_stamp_ is None:
            raise Exception("timestamp is empty!")

        return time_stamp_[0]

    def close_emited(self, *args):
        """."""
        pass
        # LOGGER.warning("** %s, %s", self, args)

    def insertMulti(
        self, table_name: str, list_records: Iterable = []
    ) -> bool:  # FIXME SQLITE NO PUEDE TODAS DE GOLPE
        """Insert several rows at once."""
        util = flutil.FLUtil()
        model_ = qsadictmodules.QSADictModules.from_project("%s_orm" % table_name)
        session_ = self.db_.connManager().dbAux().session()
        if not model_:
            return False

        util.setTotalSteps(len(list_records))  # type: ignore [arg-type]
        for number, line in enumerate(list_records):
            # model_obj = model_()
            field_names = []
            field_values = []
            for field, value in line:
                if field.generated():
                    if field.type() in ("string", "stringlist", "bytearray"):
                        value = self.normalizeValue(value)

                        if value in ["Null", "NULL"]:
                            value = "''"
                        else:
                            value = self.formatValue(field.type(), value, False)
                    else:
                        value = self.formatValue(field.type(), value, False)

                # setattr(model_obj, field.name(), value)
                field_names.append(field.name())
                field_values.append(value)

            sql = None
            if field_names:
                sql = """INSERT INTO %s(%s) values (%s)""" % (
                    table_name,
                    ", ".join(field_names),
                    ", ".join(map(str, field_values)),
                )

            if sql:
                try:
                    session_.connection().execute(sql)
                except Exception as error:
                    LOGGER.error("insertMulti: %s", str(error))
                    return False
            util.setProgress(number)
        session_.flush()
        return True

    def Mr_Proper(self) -> None:
        """Clear all garbage data."""

        util = flutil.FLUtil()
        conn_dbaux = self.db_.connManager().dbAux()

        multi_fllarge = False
        fllarge_tables_list: List[str] = []
        fllarge_to_delete: Dict[str, List[str]] = {}
        sql = "SELECT valor FROM flsettings WHERE flkey=%s" % self.formatValue(
            "string", "FLLargeMode", False
        )
        cursor = conn_dbaux.execute_query(sql)
        try:
            if cursor is not None:
                result_ = cursor.fetchone()
                if result_:
                    multi_fllarge = utils_base.text2bool(str(result_[0]))
        except Exception as error:
            LOGGER.warning("Mr_Proper: %s" % str(error))

        if not multi_fllarge:
            fllarge_tables_list.append("fllarge")

        tables = self.db_.tables("Tables")
        altered_tables = []
        for table_name in tables:
            if table_name.find("alteredtable") > -1:
                altered_tables.append(table_name)
            elif multi_fllarge and table_name.startswith("fllarge_"):
                fllarge_tables_list.append(table_name)

        LOGGER.debug(
            "Tablas fllarge. Modo multiple: %s Lista : %s sql : %s",
            multi_fllarge,
            fllarge_tables_list,
            sql,
        )

        util.createProgressDialog(
            util.translate("application", "Revisando tablas fllarge"), len(fllarge_tables_list)
        )

        for number, table_fllarge in enumerate(fllarge_tables_list):
            util.setLabelText(util.translate("application", "Revisando tabla %s" % table_fllarge))

            sql = "SELECT refkey FROM %s WHERE 1 = 1" % table_fllarge
            cursor_qry1: Iterable = conn_dbaux.execute_query(sql) or []
            old_target = ""
            metadata_target = None

            for num, line in enumerate(list(cursor_qry1)):
                target = line[0].split("@")[1]
                found = False
                if target != old_target:
                    metadata_target = self.db_.connManager().manager().metadata(target)
                    old_target = target

                if metadata_target is None:
                    LOGGER.warning("Error limpiando fllarge: %s no tiene un metdata válido", target)
                    return
                else:
                    for field in metadata_target.fieldList():
                        if found:
                            break
                        if field.type() == "pixmap":
                            sql = "SELECT %s FROM %s WHERE %s" % (
                                field.name(),
                                target,
                                "1 = 1"
                                if multi_fllarge
                                else "%s=%s"
                                % (field.name(), self.formatValue("string", line[0], False)),
                            )  # 1 a 1 , si busco especifico me da problemas mssql
                            cursor_finder: Iterable = conn_dbaux.execute_query(sql) or []
                            for result_finder in list(cursor_finder):
                                if result_finder[0] == line[0]:
                                    found = True
                                    break

                    if not found:
                        if table_fllarge not in fllarge_to_delete.keys():
                            fllarge_to_delete[table_fllarge] = []
                        fllarge_to_delete[table_fllarge].append(line[0])

            util.setProgress(number)
            QtWidgets.QApplication.processEvents()
        util.destroyProgressDialog()

        util.createProgressDialog(
            util.translate("application", "Limpiando tablas fllarge"), len(fllarge_to_delete)
        )
        for number, key in enumerate(fllarge_to_delete.keys()):
            for ref_key in fllarge_to_delete[key]:
                LOGGER.debug("Eliminado %s.%s", key, ref_key)
                util.setLabelText(util.translate("application", "Limpiando tabla fllarge %s" % key))
                sql = "DELETE FROM %s WHERE refkey = %s" % (
                    key,
                    self.formatValue("string", ref_key, False),
                )
                conn_dbaux.execute_query(sql)

            util.setProgress(number)

        util.destroyProgressDialog()

        conn_dbaux.transaction()

        # query = pnsqlquery.PNSqlQuery(None,conn_dbaux)
        cursor_qry: Iterable = (
            conn_dbaux.execute_query(
                "SELECT nombre FROM flfiles WHERE nombre "
                + self.formatValueLike("string", "%%.mtd", False)
            )
            or []
        )
        list_mtds = []
        try:
            for data in list(cursor_qry):
                list_mtds.append(data[0])
        except Exception as error:
            LOGGER.error("Mr_Proper: %s", error)

        reg_exp = re.compile("^.*\\d{6,9}$")
        bad_list_mtds = list(filter(reg_exp.match, list_mtds))

        util.createProgressDialog(
            util.translate("application", "Borrando backups"), len(bad_list_mtds)
        )

        for number, mtd_name in enumerate(bad_list_mtds):
            util.setLabelText(util.translate("application", "Borrando registro %s" % mtd_name))
            conn_dbaux.execute_query("DELETE FROM flfiles WHERE nombre ='%s'" % mtd_name)
            util.setProgress(number)

        util.setTotalSteps(len(altered_tables))

        for number, altered_table_name in enumerate(altered_tables):
            util.setLabelText(
                util.translate("application", "Borrando registro %s" % altered_table_name)
            )
            if self.existsTable(altered_table_name):
                util.setLabelText(
                    util.translate("application", "Borrando tabla %s" % altered_table_name)
                )
                conn_dbaux.execute_query(
                    "DROP TABLE %s %s" % (altered_table_name, self._text_cascade)
                )

            util.setProgress(number)

        util.destroyProgressDialog()

        tables = self.db_.tables("Tables")

        util.createProgressDialog(
            util.translate("application", "Comprobando base de datos"), len(tables) + 2
        )
        for number, table_name in enumerate(tables):
            util.setLabelText(util.translate("application", "Comprobando tabla %s" % table_name))
            metadata = conn_dbaux.connManager().manager().metadata(table_name)
            if metadata is not None:
                if self.mismatchedTable(table_name, metadata):
                    if metadata:
                        msg = util.translate(
                            "application",
                            "La estructura de los metadatos de la tabla '%s' y su "
                            "estructura interna en la base de datos no coinciden. "
                            "Intentando regenerarla." % table_name,
                        )

                        LOGGER.warning(msg)
                        self.alterTable(metadata)

            util.setProgress(number)

        util.destroyProgressDialog()
        self.checkSequences()

        util.createProgressDialog(util.translate("application", "Comprobando base de datos"), 4)
        util.setLabelText(util.translate("application", "Borrando flmetadata"))
        util.setProgress(1)
        conn_dbaux.execute_query("DELETE FROM flmetadata")
        util.setLabelText(util.translate("application", "Borrando flvar"))
        util.setProgress(2)
        conn_dbaux.execute_query("DELETE FROM flvar")
        conn_dbaux.connManager().manager().cleanupMetaData()
        conn_dbaux.commit()

        util.setLabelText(util.translate("application", "Vacunando base de datos"))
        util.setProgress(3)
        self.vacuum()

        util.setProgress(4)
        util.destroyProgressDialog()

    def vacuum(self):
        """Vacuum tables."""

        # self._connection.connection.set_isolation_level(0)
        self.execute_query("vacuum")
        # self._connection.connection.set_isolation_level(1)

    def sqlLength(self, field_name: str, size: int) -> str:
        """Return length formated."""

        return "LENGTH(%s)=%s" % (field_name, size)

    def checkSequences(self) -> None:
        """Check sequences."""

        return

    def regenTable(self, table_name: str, new_metadata: "pntablemetadata.PNTableMetaData") -> bool:
        """Regenerate tables."""

        must_alter = self.mismatchedTable(table_name, new_metadata)

        if must_alter:
            must_alter = self.alterTable(new_metadata)

            if must_alter:
                conn_dbaux = self.db_.connManager().dbAux()
                reg_exp = re.compile("^.*\\d{6,9}$")
                bad_list_tables = list(filter(reg_exp.match, self.db_.tables("Tables")))

                sql = "SELECT nombre FROM flfiles WHERE nombre%s" % self.formatValueLike(
                    "string", "%%alteredtable", False
                )
                cursor: Iterable = conn_dbaux.execute_query(sql) or []
                for name in cursor:
                    sql = "DELETE FROM flfiles WHERE nombre ='%s'" % name
                    conn_dbaux.execute_query(sql)
                    if name.find("alteredtable") > -1 and self.existsTable(
                        name.replace(".mtd", "")
                    ):
                        conn_dbaux.execute_query(
                            "DROP TABLE %s %s" % (name.replace(".mtd", ""), self._text_cascade)
                        )

                for bad_table in bad_list_tables:
                    if self.existsTable(bad_table):
                        conn_dbaux.execute_query(
                            "DROP TABLE %s %s" % (bad_table, self._text_cascade)
                        )

                conn_dbaux.commit()

        return must_alter

    @classmethod
    def close_connection_warning(cls, dbapi_connection, connection_record=None) -> None:
        """Show connection closed message."""
        if application.SHOW_CONNECTION_EVENTS:
            LOGGER.warning(
                "The connection was closed: connection: %s, record : %s !!",
                dbapi_connection,
                connection_record,
            )

    def pool_enabled(self) -> bool:
        """Return if pool is enabled."""

        mng_ = self.db_.connManager()
        return mng_.limit_connections > 0

    def get_common_params(self) -> None:
        """Load common params."""

        self._queqe_params["encoding"] = "UTF-8"

        mng_ = self.db_.connManager()
        limit_conn = mng_.limit_connections

        if self.pool_enabled():
            LOGGER.info("SqlAlchemy pool enabled")
            self._queqe_params["poolclass"] = pool.QueuePool
            self._queqe_params["pool_size"] = limit_conn
            self._queqe_params["max_overflow"] = int(limit_conn + 10)
            # if mng_.safe_mode_level in [4, 5]:
            # self._queqe_params["pool_pre_ping"] = True
            if mng_.connections_time_out:
                self._queqe_params["pool_timeout"] = int(mng_.connections_time_out)

            if self._can_use_preping and settings.CONFIG.value("ebcomportamiento/preping", False):
                self._queqe_params["pool_pre_ping"] = True

        else:
            LOGGER.debug("SqlAlchemy pool disabled")
            self._queqe_params["poolclass"] = pool.NullPool
            if mng_.connections_time_out:
                self._queqe_params["connect_args"] = {"connect_timeout": mng_.connections_time_out}

        if application.LOG_SQL:
            self._queqe_params["echo"] = True
            if limit_conn > 0:
                self._queqe_params["echo_pool"] = True

        for key, value in self._queqe_params.items():
            LOGGER.debug("    * %s = %s", key, value)

    def listen_engine(self) -> None:
        """Listen engine events."""

        event.listen(self._engine, "close_detached", self.close_connection_warning)
        event.listen(self._engine, "close", self.close_connection_warning)

    def do_begin(self, conn):
        """Begin event."""

        pass

    def do_commit(self, conn):
        """Commit event."""

        pass

    def do_rollback(self, conn):
        """Rollback event."""

        pass

    def do_savepoint(self, conn, name):
        """Save point event."""

        pass

    def do_connect(self, dbapi_connection, connection_record):
        """Isolation Level fix."""

        pass

    def invalid_engine(self, table_name: str, mute: bool = False) -> bool:
        """Return if table engine is valid."""

        return False

    def calculateChanges(
        self, table_metadata: "itablemetadata.ITableMetaData"
    ) -> Dict[str, List[Any]]:
        """Calculate changes betwen metadata and tables."""

        result: Dict[str, List] = {"upgrade": [], "downgrade": []}
        table_name = table_metadata.name()
        db_data_list = self.recordInfo2(table_metadata.name())
        meta_data_list = self.recordInfo(table_metadata)

        for db_key, db_value in db_data_list.items():
            if db_key not in meta_data_list.keys():  # drop.
                result["upgrade"].append("op.drop_column('%s', '%s')" % (table_name, db_key))

            elif self.notEqualsFields(db_value, meta_data_list[db_key], table_metadata.isQuery()):
                result["upgrade"].append(
                    self.resolveAlterColumn(
                        table_name,
                        table_metadata.field(db_key),  # type: ignore [arg-type]
                        db_value,
                    )
                )

        for field_name in meta_data_list.keys():
            if field_name not in db_data_list.keys():  # add
                col_data = self.buildColumnData(
                    table_metadata.field(field_name)  # type: ignore [arg-type]
                )
                result["upgrade"].append("op.add_column('%s', %s)" % (table_name, col_data))

        return result

    def buildColumnData(self, field_meta: "pnfieldmetadata.PNFieldMetaData") -> str:
        """Return build data."""
        data_result = pnmtdparser.generate_field(field_meta, "sa")

        def_value: Optional[str] = field_meta.defaultValue()
        text_default_value = ""
        if not field_meta.allowNull():
            text_default_value = " ,server_default='%s'" % (
                def_value
                if def_value is not None
                else self.formatValue(field_meta.type(), None, False)
            )

        result = "sa.Column('%s', %s, nullable=%s%s)" % (
            field_meta.name(),
            data_result,
            field_meta.allowNull(),
            text_default_value,
        )

        return result

    def resolveAlterColumn(
        self,
        table_name: str,
        field_meta: "pnfieldmetadata.PNFieldMetaData",
        db_value: List[Union[int, str]],
    ):
        """Return string alter column."""

        def_value: Optional[str] = field_meta.defaultValue()
        text_default_value = ""
        if not field_meta.allowNull():
            text_default_value = " ,server_default='%s'" % (
                def_value
                if def_value is not None
                else self.formatValue(field_meta.type(), None, False)
            )

        return "op.alter_column('%s', '%s', %s)" % (
            table_name,
            field_meta.name(),
            "type_=%s, existing_type=%s, nullable=%s%s"
            % (
                pnmtdparser.generate_field(field_meta, "sa"),
                pnmtdparser.resolve_type(db_value[1], db_value[3], "sa"),  # type: ignore [arg-type]
                field_meta.allowNull(),
                text_default_value,
            ),
        )



"""Test_process module."""

import unittest
from pineboolib.loader.main import init_testing, finish_testing
from pineboolib.application.packager import pnpackager
from pineboolib.application.packager.tests import fixture_path


class TestPNPAckager(unittest.TestCase):
    """TestUnpacker Class."""


class TestProcess(unittest.TestCase):
    """TestProcess Class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_pnpackager(self) -> None:
        """Test eneboopkgs load."""
        from pineboolib import application
        from pineboolib.fllegacy import systype
        from pineboolib.qsa import qsa
        import os

        file_name = "%s/package.eneboopkg" % application.PROJECT.tmpdir
        if os.path.exists(file_name):
            os.remove(file_name)
        packager = pnpackager.PNPackager(file_name)
        result = packager.pack(fixture_path("principal"))
        self.assertTrue(result, packager.errorMessages())
        self.assertTrue(os.path.exists(file_name))
        self.assertTrue(os.path.exists("%s/modules.def" % os.path.dirname(file_name)))
        self.assertTrue(os.path.exists("%s/files.def" % os.path.dirname(file_name)))

        qsa_sys = systype.SysType()
        self.assertTrue(qsa_sys.loadModules(file_name, False))

        qry = qsa.FLSqlQuery()
        qry.setTablesList("flfiles")
        qry.setSelect("count(nombre)")
        qry.setFrom("flfiles")
        qry.setWhere("1=1")
        self.assertTrue(qry.exec_())
        self.assertTrue(qry.first())
        self.assertEqual(qry.value(0), 13)

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""
        finish_testing()



"""Qlineedit module."""

# -*- coding: utf-8 -*-
from PyQt6 import QtWidgets  # type: ignore[import]
from pineboolib.core import decorators


from typing import Optional, Any, TYPE_CHECKING

if TYPE_CHECKING:
    from pineboolib.q3widgets import qframe  # noqa: F401 # pragma: no cover
    from pineboolib.q3widgets import qgroupbox  # noqa: F401 # pragma: no cover
    from pineboolib.q3widgets import qwidget  # noqa: F401 # pragma: no cover
    from pineboolib.q3widgets import qlineedit  # noqa: F401 # pragma: no cover


class QLineEdit(QtWidgets.QLineEdit):
    """QLineEdit class."""

    _parent = None
    WindowOrigin = 0
    autoSelect: bool = False

    def __init__(self, parent: Optional[Any] = None, name: Optional[str] = None) -> None:
        """Inicialize."""

        super(QLineEdit, self).__init__(parent)
        self._parent = parent

        if name:
            self.setObjectName(name)

        self.setMaximumHeight(22)

    def getText(self) -> str:
        """Return the text of the field."""

        return super().text()

    def setText(self, text: Any) -> None:
        """Set the text of the field."""

        super().setText(str(text))

    text: str = property(getText, setText)  # type: ignore [assignment] # noqa F821

    @decorators.not_implemented_warn
    def setBackgroundOrigin(self, bgo: Any):
        """Not implemented."""
        pass

    @decorators.not_implemented_warn
    def setLineWidth(self, width: int):
        """Not implemented."""
        pass

    @decorators.not_implemented_warn
    def setFrameShape(self, frame_shape: int):
        """Not implemented."""
        pass

    @decorators.not_implemented_warn
    def setFrameShadow(self, frame_shadow: int):
        """Not implemented."""
        pass



# -*- coding: utf-8 -*-
"""
Variety of tools for KUT.
"""
import os
import sys
import datetime
import fnmatch
from xml.etree.ElementTree import Element, ElementTree
from typing import Any, Iterable, Optional, List

from pineboolib import logging
from pineboolib.core.utils.utils_base import load2xml
from pineboolib.application.utils import date_conversion, xpm
from pineboolib.core import settings

from PyQt6 import QtXml, QtCore  # type: ignore[import]

from pineboolib.application.database import pnsqlquery

LOGGER = logging.get_logger(__name__)


class KParserTools(object):
    """
    Common functions for different KUT parsers.
    """

    _fix_altura = None

    def __init__(self) -> None:
        """Create base class for tools."""

        self.pagina = 0
        self._fix_ratio_h = 0.927  # Corrector de altura 0.927
        self._fix_ratio_w = 0.92

    def loadKut(self, data: str) -> "ElementTree":
        """
        Parse KUT xml from text.

        @param data. Input text (kut sources)
        @return xml.
        """
        return load2xml(data)

    def ratio_correction_h(self, value: float) -> int:
        """
        Revise height to become similar to the original kut.

        @param value. Input number to revise.
        @return Revised number.
        """
        return int(value * self._fix_ratio_h)

    def ratio_correction_w(self, value: float) -> int:
        """
        Revise width to become similar to the original kut.

        @param value. Input number to revise.
        @return Revised number.
        """
        return int(value * self._fix_ratio_w)

    def convertToNode(self, data: Element) -> "QtXml.QDomElement":
        """
        Convert XML line to Node XML.

        When it's a calculatedField sends data to QSA of Node type.
        @param data. xml with related line info.
        @return Node with original data contents.
        """

        # node = Node()

        doc = QtXml.QDomDocument()
        ele = doc.createElement("element")
        for k in data.keys():
            attr_node = doc.createAttribute(k)
            attr_node.setValue(data.get(k) or "")
            ele.setAttributeNode(attr_node)

        return ele

    def getHeight(self, xml: Element) -> int:
        """
        Retrieve height specified in xaml.

        @param xml. Element to extract the height from.
        @return Height or zero if does not exist.
        """
        return int(xml.get("Height") or "0")

    def getSpecial(self, name: str, page_num: Optional[int] = None) -> str:
        """
        Retrieve value of special type.

        @param name. Name of special type to load.
        @param page_num. PAge number if it is "PageNo" type.
        @return Required value.
        """
        LOGGER.debug("%s:getSpecial %s" % (__name__, name))
        ret = "None"
        if name:
            if name[0] == "[":
                name = name[1:-1]
            if name in ("Fecha", "Date"):
                ret = str(datetime.date.__format__(datetime.date.today(), "%d-%m-%Y"))
            elif name in ("NúmPágina", "PageNo", "NÃºmPÃ¡gina"):
                ret = str(page_num)

        return ret

    def calculated(
        self,
        value: Any,
        data_type: int,
        xml: Optional["Element"] = None,
        data: Optional["Element"] = None,
    ) -> Any:
        """
        Get value of type "calculated".

        @param field. Field name
        @param dataType. Data type. Changes how value is returned.
        @param precision. Number of decimal places.
        @param data. XML data line related.
        @return calculated value.
        """

        precision = 0
        type_ = None
        date_format_num = None

        # if value is not None:
        #    value = self.restore_text(value)

        if xml is not None:
            precision = int(xml.get("Precision") or 0)
            type_ = xml.get("Type")
            date_format_num = xml.get("DateFormat")

        if data_type in (0, 1) and data is not None:  # str
            if data_type == 1:
                value = data.get(value)
            # 0 pasa porque ya viene procesado ...

        elif data_type == 2:  # float
            if type_ is None:
                if value not in (None, "None", ""):
                    value = self.restore_text(value)
                    value = QtCore.QLocale.system().toString(float(value), "f", precision)

        elif data_type == 3:  # time
            if value.find("T") > -1:
                value = date_conversion.date_amd_to_dma(value[: value.find("T")])

            sep = "-"
            if date_format_num is not None:
                if date_format_num == "11":
                    value = value.replace(sep, ".")
                elif date_format_num == "18":
                    value = value.replace(sep, "/")
                elif date_format_num == "19":
                    value = value.replace(sep, "-")
                else:
                    LOGGER.warning("UNKNOWN DateFormat %s --> %s", date_format_num, value)
        elif data_type == 4:  # currency
            if value not in (None, "None"):
                float_value = float(value)
                if float_value < -0.01 or float_value > 0.01:
                    value = QtCore.QLocale.system().toString(float_value, "f", 2)

        elif data_type == 5:  # 5 pixmap
            pass

        elif data_type == 6:  # 6 barcode
            pass

        elif data_type == 7:  # bool
            value = "No" if str(value).upper() in ["FALSE", "F", "0"] else "Sí"

        elif data is not None:
            value = data.get(value)

        return value

    def parseKey(self, ref_key: Optional[str] = None) -> Optional[str]:
        """
        Get filename of .png file cached on tempdata. If it does not exist it is created.

        @param. String of related tuple in fllarge.
        @return. Path to the file in tempdata.
        """

        ret = None
        table_name = "fllarge"
        if ref_key is not None:
            from PyQt6.QtGui import QPixmap  # type: ignore[import]

            value = None
            tmp_dir = settings.CONFIG.value("ebcomportamiento/temp_dir")
            img_file = "%s/%s.png" % (tmp_dir, ref_key)

            if not os.path.exists(img_file) and ref_key[0:3] == "RK@":
                single_query = pnsqlquery.PNSqlQuery()
                single_query.exec_("SELECT valor FROM flsettings WHERE flkey='FLLargeMode'")
                one_fllarge = True

                if single_query.next():
                    if single_query.value(0) == "True":
                        one_fllarge = False

                if (
                    not one_fllarge
                ):  # Si no es FLLarge modo único añadimos sufijo "_nombre" a fllarge
                    table_name += "_%s" % ref_key.split("@")[1]

                qry = pnsqlquery.PNSqlQuery()
                if qry.exec_("SELECT contenido FROM %s WHERE refkey='%s'" % (table_name, ref_key)):
                    if qry.next():
                        value = xpm.cache_xpm(qry.value(0))

                if value:
                    ret = img_file
                    pix = QPixmap(value)
                    if not pix.save(img_file):
                        LOGGER.warning(
                            "%s:refkey2cache No se ha podido guardar la imagen %s"
                            % (__name__, img_file)
                        )
                        ret = None
                    else:
                        ret = img_file
            elif ref_key.endswith(".xpm"):
                pix = QPixmap(ref_key)
                img_file = ref_key.replace(".xpm", ".png")
                if not pix.save(img_file):
                    LOGGER.warning(
                        "%s:refkey2cache No se ha podido guardar la imagen %s"
                        % (__name__, img_file)
                    )
                    ret = None
                else:
                    ret = img_file

            else:
                ret = img_file

        return ret

    def convertPageSize(
        self, size: int, orientation: int, custom: Optional[List[int]] = None
    ) -> List[int]:
        """
        Get page size for the page code specified on .kut file.

        @param size. Size code specified on doc (0..31).
        @param orientation. 0 portrait, 1 landscape.
        @param custom. Where size is (30 or 31), custom is returned.
        @return Array with the size values.
        """
        result = None
        if size == 0:
            result = [595, 842]  # "A4"
        elif size == 1:
            result = [709, 499]  # "B5"
        elif size == 2:
            result = [612, 791]  # "LETTER"
        elif size == 3:
            result = [612, 1009]  # "legal"
        elif size == 5:
            result = [2384, 3370]  # "A0"
        elif size == 6:
            result = [1684, 2384]  # "A1"
        elif size == 7:
            result = [1191, 1684]  # "A2"
        elif size == 8:
            result = [842, 1191]  # "A3"
        elif size == 9:
            result = [420, 595]  # "A5"
        elif size == 10:
            result = [298, 420]  # "A6"
        elif size == 11:
            result = [210, 298]  # "A7"
        elif size == 12:
            result = [147, 210]  # "A8"
        elif size == 13:
            result = [105, 147]  # "A9"
        elif size == 14:
            result = [4008, 2835]  # "B0"
        elif size == 15:
            result = [2835, 2004]  # "B1"
        elif size == 16:
            result = [125, 88]  # "B10"
        elif size == 17:
            result = [2004, 1417]  # "B2"
        elif size == 18:
            result = [1417, 1001]  # "B3"
        elif size == 19:
            result = [1001, 709]  # "B4"
        elif size == 20:
            result = [499, 354]  # "B6"
        elif size == 21:
            result = [324, 249]  # "B7"
        elif size == 22:
            result = [249, 176]  # "B8"
        elif size == 23:
            result = [176, 125]  # "B9"
        elif size == 24:
            result = [649, 459]  # "C5"
        elif size == 25:
            result = [113, 79]  # "C10"
        elif size == 28:
            result = [1255, 791]  # "LEDGER"
        elif size == 29:
            result = [791, 1255]  # "TABLOID"
        elif size in (30, 31):
            result = custom  # "CUSTOM"
        if result is None:
            LOGGER.warning("porcessXML:No se encuentra pagesize para %s. Usando A4" % size)
            result = [595, 842]

        # if orientation != 0:
        #    r = [r[1], r[0]]

        return result

    def find_font(self, font_name: str, font_style: str) -> Optional[str]:
        """
        Find and retrieve path for a specified font.

        @param font_name. Font name required
        @return Path to ".ttf" or None
        """
        fonts_folders: List[str] = []
        if sys.platform.find("darwin") > -1:
            fonts_folders = ["/Library/Fonts", "/System/Library/Fonts", "~/Library/Fonts"]

        elif sys.platform.find("win") > -1:
            windirs = os.environ.get("WINDIR")
            if windirs is None:
                raise Exception("WINDIR environ not found!")

            for win_dir in windirs.split(";"):
                fonts_folders.append(os.path.join(win_dir, "fonts"))

        elif sys.platform.find("linux") > -1:
            lindirs = os.environ.get("XDG_DATA_DIRS", "")
            if not lindirs:
                lindirs = "usr/share"

            for lin_dir in lindirs.split(":"):
                fonts_folders.append(os.path.join(lin_dir, "fonts"))
        else:
            LOGGER.warning("KUTPARSERTOOLS: Plataforma desconocida %s", sys.platform)
            return None

        font_name = font_name.replace(" ", "_")

        font_name = font_name
        font_name2 = font_name
        font_name3 = font_name

        if font_name.endswith("BI"):
            font_name2 = font_name.replace("BI", "_Bold_Italic")
            font_name3 = font_name.replace("BI", "bi")

        if font_name.endswith("B"):
            font_name2 = font_name.replace("B", "_Bold")
            font_name3 = font_name.replace("B", "b")

        if font_name.endswith("I"):
            font_name2 = font_name.replace("I", "_Italic")
            font_name3 = font_name.replace("I", "i")

        for folder in fonts_folders:
            for root, dirnames, filenames in os.walk(folder):
                for filename in fnmatch.filter(filenames, "%s.ttf" % font_name):
                    ret_ = os.path.join(root, filename)
                    return ret_

                for filename in fnmatch.filter(
                    filenames, "%s%s.ttf" % (font_name[0].upper(), font_name[1:])
                ):
                    ret_ = os.path.join(root, filename)
                    return ret_

                for filename in fnmatch.filter(filenames, "%s.ttf" % font_name2):
                    ret_ = os.path.join(root, filename)
                    return ret_

                for filename in fnmatch.filter(
                    filenames, "%s%s.ttf" % (font_name2[0].upper(), font_name2[1:])
                ):
                    ret_ = os.path.join(root, filename)
                    return ret_

                for filename in fnmatch.filter(filenames, "%s.ttf" % font_name3):
                    ret_ = os.path.join(root, filename)
                    return ret_

                for filename in fnmatch.filter(
                    filenames, "%s%s.ttf" % (font_name3[0].upper(), font_name3[1:])
                ):
                    ret_ = os.path.join(root, filename)
                    return ret_

        return None

    def calculate_sum(self, field_name: str, line: Element, xml_list: Iterable, level: int) -> str:
        """
        Calculate sum for specified element line. level equal or upper.
        """
        val = 0.0
        min_level = -1
        last_level = -1
        for item in xml_list:
            lev_ = int(item.get("level"))
            val_ = float(item.get(field_name))

            if min_level < lev_ or lev_ < last_level:
                # print("nuevo min_level", lev_, "previo", min_level)
                min_level = lev_
                val = 0.0

            if min_level >= level and lev_ >= level:
                val += val_

            # print(
            #    "line_level:",
            #    lev_,
            #    "line_val:",
            #    val_,
            #    "obj_level:",
            #    level,
            #    "current_val:",
            #    val,
            #    "min_level:",
            #    min_level,
            # )
            if item is line:
                break

            last_level = lev_

        return str(val)

    def restore_text(self, text: str) -> str:
        """Un-replace text for special characters."""
        ret_ = text
        ret_ = ret_.replace("__RPAREN__", ")")
        ret_ = ret_.replace("__LPAREN__", "(")
        ret_ = ret_.replace("__ASTERISK__", "*")
        ret_ = ret_.replace("__PLUS__", "+")
        ret_ = ret_.replace("__MINUS__", "-")
        ret_ = ret_.replace("__BLANK__", " ")

        return ret_



"""DGI Module."""

from pineboolib import logging

from typing import Callable, Any, TYPE_CHECKING

if TYPE_CHECKING:
    from pineboolib.interfaces.dgi_schema import dgi_schema  # pragma: no cover
    from pineboolib.plugins.dgi.dgi_qt import dgi_qt  # pragma: no cover
    from pineboolib.plugins.dgi.dgi_fcgi import dgi_fcgi  # pragma: no cover


LOGGER = logging.get_logger(__name__)


def load_dgi(name: str, param: Any) -> "dgi_schema":
    """Load a DGI module dynamically."""

    dgi_entrypoint = DGILoader.load_dgi(name)

    try:
        dgi = dgi_entrypoint()  # FIXME: Necesitamos ejecutar código dinámico tan pronto?
    except Exception:
        LOGGER.exception("Error inesperado al cargar el módulo DGI %s" % name)
        raise

    if param:
        dgi.setParameter(param)

    LOGGER.debug("DGI loaded: %s", name)

    return dgi


class DGILoader(object):
    """DGILoader Class."""

    @staticmethod
    def load_dgi_qt() -> "dgi_qt.DgiQt":
        """Load dgi qt."""

        from pineboolib.plugins.dgi.dgi_qt import dgi_qt as dgi

        return dgi.DgiQt()

    @staticmethod
    def load_dgi_fcgi() -> "dgi_fcgi.DgiFcgi":
        """Load dgi fcgi."""

        from pineboolib.plugins.dgi.dgi_fcgi import dgi_fcgi as dgi

        return dgi.DgiFcgi()

    @classmethod
    def load_dgi(cls, name: str) -> Callable:
        """Load dgi specified by name."""

        loader = getattr(cls, "load_dgi_%s" % name, None)
        if not loader:
            raise ValueError("Unknown DGI %s" % name)
        return loader



"""Flworkspace module."""
# -*- coding: utf-8 -*-
from pineboolib.fllegacy.flwidget import FLWidget
from typing import Any


class FLWorkSpace(FLWidget):
    """FLWorkSpace class."""

    def __getattr__(self, name: str) -> Any:
        """Return a parent attribute."""

        return getattr(self.parent(), name)



"""Load_script module."""

from importlib.machinery import ModuleSpec
from pineboolib.core.utils import logging
from pineboolib.application.utils.path import _path

from typing import Optional, TYPE_CHECKING

from pineboolib.application.staticloader import pnmodulesstaticloader
from pineboolib.application.parsers.parser_mtd import pnmtdparser
from pineboolib import application

import xml.etree.ElementTree as ET
from importlib import util
from sqlalchemy.ext import declarative  # type: ignore [import]
from sqlalchemy import exc  # type: ignore [import]

import shutil
import time
import os

if TYPE_CHECKING:
    from pineboolib.qsa import formdbwidget  # pragma: no cover
    from pineboolib.application import xmlaction  # pragma: no cover
    from types import ModuleType  # pragma: no cover

LOGGER = logging.get_logger(__name__)


def load_script(script_name: str, action_: "xmlaction.XMLAction") -> "formdbwidget.FormDBWidget":
    """
    Transform QS script into Python and starts it up.
    """

    # LOGGER.info(
    #    "LOADING SCRIPT %s (ALWAYS PARSE QSA: %s) ----->",
    #    script_name.upper(),
    #    application.PROJECT.no_python_cache,
    # )

    script_name = script_name.replace(".qs", "")
    LOGGER.debug("Loading script %s for action %s", script_name, action_._name)

    script_path_py: str = ""
    script_path_qs: str = ""
    script_loaded = None

    cached_script_path_qs: str = _path("%s.qs" % script_name, False) or ""
    # Busqueda en carpetas .py
    cached_script_path_py: str = _path("%s.py" % script_name, False) or ""
    if cached_script_path_qs and not cached_script_path_py:  # busqueda en carpetas .qs.py
        file_py = "%spy" % cached_script_path_qs[:-2]
        cached_script_path_py = file_py if os.path.exists(file_py) else ""

    if application.PROJECT.no_python_cache and cached_script_path_qs:  # Si no_python_cache
        cached_script_path_py = ""

    # carga estática
    static_flag = (
        "%s/static.xml" % os.path.dirname(cached_script_path_qs) if cached_script_path_qs else ""
    )
    script_path_py_static = _static_file("%s.py" % script_name)
    script_path_qs_static = _static_file("%s.qs" % script_name) if not script_path_py_static else ""

    # Primera opción carga estática.
    if script_path_py_static:
        script_path_py = script_path_py_static
    elif script_path_qs_static:
        script_path_qs = script_path_qs_static
    else:  # Segunda Caché
        if cached_script_path_py:
            script_path_py = cached_script_path_py
        elif cached_script_path_qs:
            script_path_qs = cached_script_path_qs

    if script_path_py:
        _remove(static_flag)

        if script_path_py_static and cached_script_path_py.find("system_module") == -1:
            if not cached_script_path_py:
                msg = (
                    "The %s.py file that does not exist in pineboo's cache is being overloaded."
                    % script_name
                )
                msg += "Check that the file exists in the flfiles table"
                LOGGER.exception(msg)
            # si es carga estática y no es módulo sistema lo marco
            static_flag = "%s/static.xml" % os.path.dirname(cached_script_path_py)
            _build_static_flag(static_flag, cached_script_path_py, script_path_py)

        if not os.path.isfile(script_path_py):
            raise IOError
        try:
            script_loaded = _load(script_name, script_path_py, False)
        except Exception:
            LOGGER.exception("ERROR al cargar script PY para la accion %s:", action_._name)

    elif script_path_qs:
        if not os.path.isfile(script_path_qs):
            raise IOError

        need_parse = True
        script_path_py = "%spy" % cached_script_path_qs[:-2]  # Buscamos

        if script_path_qs_static:
            replace_static = True
            if os.path.exists(static_flag):
                replace_static = _resolve_flag(
                    static_flag, cached_script_path_qs, script_path_qs_static
                )
            if replace_static:
                shutil.copy(script_path_qs_static, cached_script_path_qs)  # Lo copiamos en tempdata
                # _remove(script_path_py)
                _build_static_flag(static_flag, cached_script_path_qs, script_path_qs_static)
            else:
                need_parse = not os.path.exists(script_path_py)
        else:
            if not application.PROJECT.no_python_cache and os.path.exists(script_path_py):
                need_parse = False

        if need_parse:
            _remove(script_path_py)
            application.PROJECT.message_manager().send(
                "status_help_msg", "send", ["Convirtiendo script... %s" % script_name]
            )

            LOGGER.info(
                "PARSE_SCRIPT (name : %s, use cache : %s, file: %s",
                script_name,
                not application.PROJECT.no_python_cache,
                cached_script_path_qs,
            )
            if not application.PROJECT.parse_script_list([cached_script_path_qs]):
                if not os.path.exists(script_path_py):
                    raise Exception("El fichero %s no se ha podido crear\n" % script_path_py)

        try:
            script_loaded = _load(script_name, script_path_py, False)
        except Exception as error:
            _remove(script_path_py)
            _remove(static_flag)

            raise Exception(
                "ERROR al cargar script QS para la accion %s: %s" % (action_._name, str(error))
            )

    return _init_internal_obj(action_, script_loaded)


def _init_internal_obj(
    action_: "xmlaction.XMLAction", script_loaded: Optional["ModuleType"] = None
) -> "formdbwidget.FormDBWidget":
    """Return formDBWidget."""

    if script_loaded is None:
        from pineboolib.qsa import emptyscript

        script_loaded = emptyscript

    return script_loaded.FormInternalObj(action_)  # type: ignore[attr-defined] # noqa: F821


def _resolve_flag(static_flag: str, script: str, script_static: str) -> bool:
    """Return if replace."""

    result = True
    try:
        tree = ET.parse(static_flag)
        root = tree.getroot()
        if root.get("path_legacy") != script:
            pass
        elif root.get("last_modified_remote") != str(time.ctime(os.path.getmtime(script_static))):
            pass
        else:
            result = False
    except Exception:
        flag_file = open(static_flag, "r", encoding="UTF8")
        flag_data = flag_file.read()
        flag_file.close()
        LOGGER.warning(
            "A problem found reading %s data: %s. Forcing realoading", static_flag, flag_data
        )

    return result


def _build_static_flag(flag: str, script: str, static: str) -> None:
    """Build flag file."""

    _remove(flag)

    xml_data = get_static_flag(script, static)
    my_data = ET.tostring(xml_data, encoding="utf8", method="xml")

    file_ = open(flag, "wb")
    file_.write(my_data)
    file_.close()


def _remove(file_name: str) -> None:
    """Remove file."""

    if os.path.exists(file_name):
        try:
            os.remove(file_name)
        except FileNotFoundError as error:  # noqa: F841
            LOGGER.warning("File %s exists but not found!!", file_name)


def load_model(script_name: str, script_path_py: str) -> Optional["type"]:
    """Return model_class from path."""

    model_class: Optional["type"] = None
    script_path_py = _resolve_script("%s_model.py" % script_name, script_path_py)
    if script_path_py:
        if pnmtdparser.use_mtd_fields(script_path_py):
            script_path_py = pnmtdparser.populate_fields(script_path_py, "%s.mtd" % script_name)
            LOGGER.info(
                "El model %s no contenía legacy_metadata. Se rellena con datos de %s.mtd",
                script_name,
                script_name,
            )

        class_name = "%s%s" % (script_name[0].upper(), script_name[1:])
        script_loaded = _load("model.%s" % class_name, script_path_py)
        module_class = getattr(script_loaded, class_name, None)
        if module_class:
            module_class.__metaclass__ = "Base"
            try:
                model_class = type(class_name, (module_class, declarative.declarative_base()), {})
            except exc.ArgumentError:
                LOGGER.warning(
                    "Error in %s model. Please check columns and make sure exists a primaryKey column"
                    % script_name
                )
        else:
            LOGGER.warning(
                "No existe la clase %s dentro de %s. No se puede usar este orm."
                % (class_name, "%s_model.py" % script_name)
            )

    return model_class


def load_class(script_name: str):
    """Return class from path."""

    class_loaded = None
    script_path_py = _resolve_script("%s.py" % script_name)

    if script_path_py:
        class_name = "%s%s" % (script_name[0].upper(), script_name[1:])
        script_loaded = _load(script_name, script_path_py)
        try:
            class_loaded = getattr(script_loaded, class_name, None)
        except Exception as error:
            LOGGER.error("Error loading class %s: %s", script_name, str(error))

    return class_loaded


def load_module(script_name: str) -> Optional["ModuleType"]:
    """Return class from path."""

    script_path_py = _resolve_script(script_name)

    return _load(script_name[:-3], script_path_py) if script_path_py else None


def _resolve_script(file_name, alternative: str = "") -> str:
    """Resolve script."""

    result = _static_file(file_name)
    if not result:
        result = alternative
        if not result:
            result = _path(file_name, False) or ""

    return result or ""


def _load(  # type: ignore [return] # noqa: F821, F723
    module_name: str, script_name: str, capture_error: bool = True
) -> "ModuleType":
    """Load modules."""

    if os.path.exists(script_name) and not os.access(script_name, os.R_OK):
        if capture_error:
            LOGGER.error("Permision denied for read %s file", script_name)
        else:
            raise PermissionError

    return import_path(module_name, script_name, capture_error)


def import_path(module_name: str, script_name: str, capture_error: bool = True) -> "ModuleType":
    """Import path."""

    try:
        spec: Optional["ModuleSpec"] = util.spec_from_file_location(module_name, script_name)
        if spec and spec.loader is not None:
            module = util.module_from_spec(spec)
            spec.loader.exec_module(module)  # type: ignore [attr-defined]
            return module
        else:
            raise Exception("Module named %s can't be loaded from %s" % (module_name, script_name))

    except Exception as error:
        if capture_error:
            LOGGER.error("Error loading module %s: %s", script_name, str(error), stack_info=True)
        else:
            raise error


def _static_file(file_name: str) -> str:
    """Return static file."""

    result = ""
    mng_modules = application.PROJECT.conn_manager.managerModules()
    if mng_modules.static_db_info_ and mng_modules.static_db_info_.enabled_:
        result = pnmodulesstaticloader.PNStaticLoader.content(
            file_name, mng_modules.static_db_info_, True
        )  # Con True solo devuelve el path

    return result


def get_static_flag(database_path: str, static_path: str) -> "ET.Element":
    """Return static_info."""

    xml_data = ET.Element("data")
    xml_data.set("path_legacy", database_path)
    xml_data.set("path_remote", static_path)
    xml_data.set("last_modified_remote", str(time.ctime(os.path.getmtime(static_path))))
    return xml_data



"""Flformrecord module."""
# -*- coding: utf-8 -*-


from PyQt6 import QtCore, QtGui, QtWidgets  # type: ignore[import]

from pineboolib.core.utils import utils_base
from pineboolib.core import settings, decorators

from pineboolib.application.database import pnsqlcursor
from pineboolib import logging, application


from pineboolib.fllegacy import flformdb, flfielddb

from typing import cast, Union, Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from pineboolib.application.metadata import pnaction  # pragma: no cover
    from pineboolib.interfaces import isqlcursor  # noqa : F401 # pragma: no cover

DEBUG = False

LOGGER = logging.get_logger("FLFormRecordDB")


class FLFormRecordDB(flformdb.FLFormDB):
    """
    FLFormRecordDBInterface Class.

    FLFormDB subclass designed to edit records.

    Basically this class does the same as its class
    FLFormDB base, the only thing you add is two buttons
    Accept and / or Cancel to confirm or cancel
    the changes that are made to the components of
    data it contains.

    This class is suitable for loading forms
    editing records defined in metadata
    (FLTableMetaData).

    @author InfoSiAL S.L.
    """

    """
    Boton Aceptar
    """

    pushButtonAccept: Optional["QtWidgets.QToolButton"]

    """
    Boton Aceptar y continuar
    """
    pushButtonAcceptContinue: Optional["QtWidgets.QToolButton"]

    """
    Boton Primero
    """
    pushButtonFirst: Optional["QtWidgets.QToolButton"]

    """
    Boton Anterior
    """
    pushButtonPrevious: Optional["QtWidgets.QToolButton"]

    """
    Boton Siguiente
    """
    pushButtonNext: Optional["QtWidgets.QToolButton"]

    """
    Boton Ultimo
    """
    pushButtonLast: Optional["QtWidgets.QToolButton"]

    """
    Indica si se debe mostrar el botón Aceptar y Continuar
    """
    _show_accept_continue: bool

    """
    Indica que se está intentando aceptar los cambios
    """
    accepting: bool

    """
    Modo en el que inicialmente está el cursor
    """
    _initial_mode_access: int

    """
    Registra el nivel de anidamiento de transacciones en el que se entra al iniciar el formulario
    """
    _init_translation_level: int

    def __init__(
        self,
        action: "pnaction.PNAction",
        parent_or_cursor: Optional[Union["QtWidgets.QWidget", "isqlcursor.ISqlCursor", int]] = None,
        load: bool = False,
    ) -> None:
        """
        Inicialize.
        """
        LOGGER.trace(
            "__init__: parent_or_cursor=%s, action=%s, load=%s",
            parent_or_cursor,
            action,
            load,
        )

        cursor: Optional["pnsqlcursor.PNSqlCursor"] = None
        parent: Optional["QtWidgets.QWidget"] = None

        if isinstance(parent_or_cursor, pnsqlcursor.PNSqlCursor):
            cursor = parent_or_cursor
        elif isinstance(parent_or_cursor, QtWidgets.QWidget):
            parent = parent_or_cursor

        super().__init__(action, parent, load)

        self.setWindowModality(QtCore.Qt.WindowModality.ApplicationModal)

        if cursor:
            self.setCursor(cursor)
        LOGGER.trace("__init__: load formRecord")
        self._ui_name = action.formRecord()
        self.pushButtonAccept = None  # pylint: disable=invalid-name
        self.pushButtonAcceptContinue = None  # pylint: disable=invalid-name
        self.pushButtonFirst = None  # pylint: disable=invalid-name
        self.pushButtonPrevious = None  # pylint: disable=invalid-name
        self.pushButtonNext = None  # pylint: disable=invalid-name
        self.pushButtonLast = None  # pylint: disable=invalid-name

        self.accepting = False
        self._show_accept_continue = True
        self._initial_mode_access = pnsqlcursor.PNSqlCursor.Browse

        if self.cursor_:
            self._initial_mode_access = self.cursor_.modeAccess()

        LOGGER.trace("__init__: load form")
        self.load()
        LOGGER.trace("__init__: init form")
        self.initForm()
        LOGGER.trace("__init__: done")
        self.loop = False

    def setCaptionWidget(self, text: Optional[str] = None) -> None:
        """
        Set the window title.
        """
        if not self.cursor_:
            return

        if not text:
            text = self.cursor_.metadata().alias()

        try:
            if self.cursor_.modeAccess() == self.cursor_.Insert:
                self.setWindowTitle("Insertar %s" % text)
            elif self.cursor_.modeAccess() == self.cursor_.Edit:
                self.setWindowTitle("Editar %s" % text)
            elif self.cursor_.modeAccess() == self.cursor_.Browse:
                self.setWindowTitle("Visualizar %s" % text)
        except RuntimeError as error:
            LOGGER.warning(str(error))

    def formClassName(self) -> str:
        """
        Return the class name of the form at runtime.
        """

        return "FormRecordDB"

    def initForm(self) -> None:
        """
        Initialize the form.
        """

        if self.cursor_ and self.cursor_.metadata():
            # caption = None
            if self._action:
                self.cursor().setAction(self._action)
                if self._action.description():
                    self.setWhatsThis(self._action.description())
                self._id_mdi = self._action.name()

            # self.bindIface()
            # self.setCursor(self.cursor_)

        else:
            self.setCaptionWidget("No hay metadatos")

        acl = application.PROJECT.aq_app.acl()
        if acl:
            acl.process(self)

    def loadControls(self) -> None:
        """Load widgets for this form."""
        if self.pushButtonAcceptContinue:
            self.pushButtonAcceptContinue.hide()

        if self.pushButtonAccept:
            self.pushButtonAccept.hide()

        if self.pushButtonCancel:
            self.pushButtonCancel.hide()

        if self.pushButtonFirst:
            self.pushButtonFirst.hide()

        if self.pushButtonPrevious:
            self.pushButtonPrevious.hide()

        if self.pushButtonNext:
            self.pushButtonNext.hide()

        if self.pushButtonLast:
            self.pushButtonLast.hide()

        if self.bottomToolbar and self.toolButtonClose:
            self.toolButtonClose.hide()

        self.init_tool_bar()

        size_policy = QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Policy(0), QtWidgets.QSizePolicy.Policy(0)
        )
        size_policy.setHeightForWidth(True)

        push_button_size = self._icon_size

        if settings.CONFIG.value("application/isDebuggerMode", False):
            pushButtonExport = QtWidgets.QToolButton()  # pylint: disable=invalid-name
            pushButtonExport.setObjectName("pushButtonExport")
            pushButtonExport.setSizePolicy(size_policy)
            pushButtonExport.setMinimumSize(push_button_size)
            pushButtonExport.setMaximumSize(push_button_size)
            pushButtonExport.setIcon(
                QtGui.QIcon(utils_base.filedir("./core/images/icons", "gtk-properties.png"))
            )
            pushButtonExport.setShortcut(QtGui.QKeySequence(self.tr("F3")))
            pushButtonExport.setWhatsThis("Exportar a XML(F3)")
            pushButtonExport.setToolTip("Exportar a XML(F3)")
            pushButtonExport.setFocusPolicy(QtCore.Qt.FocusPolicy.NoFocus)
            self.bottomToolbar.layout().addWidget(pushButtonExport)  # type: ignore [union-attr]
            pushButtonExport.clicked.connect(  # type: ignore [attr-defined] # noqa: F821
                self.exportToXml
            )

            if settings.CONFIG.value("ebcomportamiento/show_snaptshop_button", False):
                push_button_snapshot = QtWidgets.QToolButton()
                push_button_snapshot.setObjectName("pushButtonSnapshot")
                push_button_snapshot.setSizePolicy(size_policy)
                push_button_snapshot.setMinimumSize(push_button_size)
                push_button_snapshot.setMaximumSize(push_button_size)
                push_button_snapshot.setIcon(
                    QtGui.QIcon(utils_base.filedir("./core/images/icons", "gtk-paste.png"))
                )
                push_button_snapshot.setShortcut(QtGui.QKeySequence(self.tr("F8")))
                push_button_snapshot.setWhatsThis("Capturar pantalla(F8)")
                push_button_snapshot.setToolTip("Capturar pantalla(F8)")
                push_button_snapshot.setFocusPolicy(QtCore.Qt.FocusPolicy.NoFocus)
                self.bottomToolbar.layout().addWidget(push_button_snapshot)  # type: ignore [union-attr]
                push_button_snapshot.clicked.connect(  # type: ignore [attr-defined] # noqa: F821
                    self.saveSnapShot
                )

            spacer = QtWidgets.QSpacerItem(
                20,
                20,
                QtWidgets.QSizePolicy.Policy.Fixed,
                QtWidgets.QSizePolicy.Policy.Fixed,
            )
            self.bottomToolbar.layout().addItem(spacer)  # type: ignore [union-attr]

        if self.cursor().modeAccess() in (self.cursor().Edit, self.cursor().Browse):
            if not self.pushButtonFirst:
                self.pushButtonFirst = QtWidgets.QToolButton()
                self.pushButtonFirst.setObjectName("pushButtonFirst")
                self.pushButtonFirst.setIcon(
                    QtGui.QIcon(utils_base.filedir("./core/images/icons", "gtk-goto-first-ltr.png"))
                )
                self.pushButtonFirst.clicked.connect(  # type: ignore [attr-defined] # noqa: F821
                    self.firstRecord
                )
                self.pushButtonFirst.setSizePolicy(size_policy)
                self.pushButtonFirst.setMaximumSize(push_button_size)
                self.pushButtonFirst.setMinimumSize(push_button_size)
                self.pushButtonFirst.setShortcut(QtGui.QKeySequence(self.tr("F5")))
                self.pushButtonFirst.setWhatsThis(
                    "Aceptar los cambios e ir al primer registro (F5)"
                )
                self.pushButtonFirst.setToolTip("Aceptar los cambios e ir al primer registro (F5)")
                self.pushButtonFirst.setFocusPolicy(QtCore.Qt.FocusPolicy.NoFocus)
                self.bottomToolbar.layout().addWidget(self.pushButtonFirst)  # type: ignore [union-attr]
                # self.pushButtonFirst.show()

            if not self.pushButtonPrevious:
                self.pushButtonPrevious = QtWidgets.QToolButton()
                self.pushButtonPrevious.setObjectName("pushButtonPrevious")
                self.pushButtonPrevious.setIcon(
                    QtGui.QIcon(utils_base.filedir("./core/images/icons", "gtk-go-back-ltr.png"))
                )
                self.pushButtonPrevious.clicked.connect(  # type: ignore [attr-defined] # noqa: F821
                    self.previousRecord
                )
                self.pushButtonPrevious.setSizePolicy(size_policy)
                self.pushButtonPrevious.setMaximumSize(push_button_size)
                self.pushButtonPrevious.setMinimumSize(push_button_size)
                self.pushButtonPrevious.setShortcut(QtGui.QKeySequence(self.tr("F6")))
                self.pushButtonPrevious.setWhatsThis(
                    "Aceptar los cambios e ir al registro anterior (F6)"
                )
                self.pushButtonPrevious.setToolTip(
                    "Aceptar los cambios e ir al registro anterior (F6)"
                )
                self.pushButtonPrevious.setFocusPolicy(QtCore.Qt.FocusPolicy.NoFocus)
                self.bottomToolbar.layout().addWidget(self.pushButtonPrevious)  # type: ignore [union-attr]
                # self.pushButtonPrevious.show()

            if not self.pushButtonNext:
                self.pushButtonNext = QtWidgets.QToolButton()
                self.pushButtonNext.setObjectName("pushButtonNext")
                self.pushButtonNext.setIcon(
                    QtGui.QIcon(utils_base.filedir("./core/images/icons", "gtk-go-back-rtl.png"))
                )
                self.pushButtonNext.clicked.connect(  # type: ignore [attr-defined] # noqa: F821
                    self.nextRecord
                )
                self.pushButtonNext.setSizePolicy(size_policy)
                self.pushButtonNext.setMaximumSize(push_button_size)
                self.pushButtonNext.setMinimumSize(push_button_size)
                self.pushButtonNext.setShortcut(QtGui.QKeySequence(self.tr("F7")))
                self.pushButtonNext.setWhatsThis(
                    "Aceptar los cambios e ir al registro siguiente (F7)"
                )
                self.pushButtonNext.setToolTip(
                    "Aceptar los cambios e ir al registro siguiente (F7)"
                )
                self.pushButtonNext.setFocusPolicy(QtCore.Qt.FocusPolicy.NoFocus)
                self.bottomToolbar.layout().addWidget(self.pushButtonNext)  # type: ignore [union-attr]
                # self.pushButtonNext.show()

            if not self.pushButtonLast:
                self.pushButtonLast = QtWidgets.QToolButton()
                self.pushButtonLast.setObjectName("pushButtonLast")
                self.pushButtonLast.setIcon(
                    QtGui.QIcon(utils_base.filedir("./core/images/icons", "gtk-goto-last-ltr.png"))
                )
                self.pushButtonLast.clicked.connect(  # type: ignore [attr-defined] # noqa: F821
                    self.lastRecord
                )
                self.pushButtonLast.setSizePolicy(size_policy)
                self.pushButtonLast.setMaximumSize(push_button_size)
                self.pushButtonLast.setMinimumSize(push_button_size)
                self.pushButtonLast.setShortcut(QtGui.QKeySequence(self.tr("F8")))
                self.pushButtonLast.setWhatsThis("Aceptar los cambios e ir al último registro (F8)")
                self.pushButtonLast.setToolTip("Aceptar los cambios e ir al último registro (F8)")
                self.pushButtonLast.setFocusPolicy(QtCore.Qt.FocusPolicy.NoFocus)
                self.bottomToolbar.layout().addWidget(self.pushButtonLast)  # type: ignore [union-attr]
                # self.pushButtonLast.show()

        if not self.cursor().modeAccess() == self.cursor().Browse:
            self.pushButtonAcceptContinue = QtWidgets.QToolButton()
            self.pushButtonAcceptContinue.setObjectName("pushButtonAcceptContinue")
            self.pushButtonAcceptContinue.clicked.connect(  # type: ignore [attr-defined] # noqa: F821
                self.acceptContinue
            )
            self.pushButtonAcceptContinue.setSizePolicy(size_policy)
            self.pushButtonAcceptContinue.setMaximumSize(push_button_size)
            self.pushButtonAcceptContinue.setMinimumSize(push_button_size)
            self.pushButtonAcceptContinue.setIcon(
                QtGui.QIcon(utils_base.filedir("./core/images/icons", "gtk-refresh.png"))
            )
            self.pushButtonAcceptContinue.setShortcut(QtGui.QKeySequence(self.tr("F9")))
            self.pushButtonAcceptContinue.setWhatsThis(
                "Aceptar los cambios y continuar con la edición de un nuevo registro (F9)"
            )
            self.pushButtonAcceptContinue.setToolTip(
                "Aceptar los cambios y continuar con la edición de un nuevo registro (F9)"
            )
            self.pushButtonAcceptContinue.setFocusPolicy(QtCore.Qt.FocusPolicy.NoFocus)
            self.bottomToolbar.layout().addWidget(self.pushButtonAcceptContinue)  # type: ignore [union-attr]
            if not self._show_accept_continue:
                self.pushButtonAcceptContinue.close()
                # self.pushButtonAcceptContinue.show()

            if not self.pushButtonAccept:
                self.pushButtonAccept = QtWidgets.QToolButton()
                self.pushButtonAccept.setObjectName("pushButtonAccept")
                self.pushButtonAccept.clicked.connect(  # type: ignore [attr-defined] # noqa: F821
                    self.accept
                )

            self.pushButtonAccept.setSizePolicy(size_policy)
            self.pushButtonAccept.setMaximumSize(push_button_size)
            self.pushButtonAccept.setMinimumSize(push_button_size)
            self.pushButtonAccept.setIcon(
                QtGui.QIcon(utils_base.filedir("./core/images/icons", "gtk-save.png"))
            )
            self.pushButtonAccept.setShortcut(QtGui.QKeySequence(self.tr("F10")))
            self.pushButtonAccept.setWhatsThis("Aceptar los cambios y cerrar formulario (F10)")
            self.pushButtonAccept.setToolTip("Aceptar los cambios y cerrar formulario (F10)")
            self.pushButtonAccept.setFocusPolicy(QtCore.Qt.FocusPolicy.NoFocus)
            self.bottomToolbar.layout().addWidget(self.pushButtonAccept)  # type: ignore [union-attr]
            # self.pushButtonAccept.show()

        if not self.pushButtonCancel:
            self.pushButtonCancel = QtWidgets.QToolButton()
            self.pushButtonCancel.setObjectName("pushButtonCancel")
            try:
                self.cursor().autoCommit.connect(  # type: ignore [attr-defined] # noqa: F821
                    self.disablePushButtonCancel
                )
            except Exception:
                pass

            self.pushButtonCancel.clicked.connect(  # type: ignore [attr-defined] # noqa: F821
                self.reject
            )

        self.pushButtonCancel.setSizePolicy(size_policy)
        self.pushButtonCancel.setMaximumSize(push_button_size)
        self.pushButtonCancel.setMinimumSize(push_button_size)
        self.pushButtonCancel.setShortcut(QtGui.QKeySequence(self.tr("Esc")))
        self.pushButtonCancel.setIcon(
            QtGui.QIcon(utils_base.filedir("./core/images/icons", "gtk-stop.png"))
        )
        if not self.cursor().modeAccess() == self.cursor().Browse:
            self.pushButtonCancel.setFocusPolicy(QtCore.Qt.FocusPolicy.NoFocus)
            self.pushButtonCancel.setWhatsThis("Cancelar los cambios y cerrar formulario (Esc)")
            self.pushButtonCancel.setToolTip("Cancelar los cambios y cerrar formulario (Esc)")
        else:
            self.pushButtonCancel.setFocusPolicy(QtCore.Qt.FocusPolicy.StrongFocus)
            self.pushButtonCancel.setFocus()
            # pushButtonCancel->setAccel(4096); FIXME
            self.pushButtonCancel.setWhatsThis("Aceptar y cerrar formulario (Esc)")
            self.pushButtonCancel.setToolTip("Aceptar y cerrar formulario (Esc)")

        # pushButtonCancel->setDefault(true);
        self.bottomToolbar.layout().addItem(  # type: ignore [union-attr]
            QtWidgets.QSpacerItem(
                20,
                20,
                QtWidgets.QSizePolicy.Policy.Fixed,
                QtWidgets.QSizePolicy.Policy.Fixed,
            )
        )
        self.bottomToolbar.layout().addWidget(self.pushButtonCancel)  # type: ignore [union-attr]
        # self.pushButtonAccept.show()

        self.setFocusPolicy(QtCore.Qt.FocusPolicy.NoFocus)

        # self.toolButtonAccept = QtGui.QToolButton()
        # self.toolButtonAccept.setIcon(QtGui.QIcon(utils_base.filedir("./core/images/icons","gtk-add.png")))
        # self.toolButtonAccept.clicked.connect(self.validateForm)
        # self.bottomToolbar.layout.addWidget(self.toolButtonAccept)
        self.inicializeControls()

    def formName(self) -> str:
        """
        Return internal form name.
        """

        return "formRecord%s" % self._id_mdi

    def closeEvent(self, event: Optional["QtGui.QCloseEvent"]) -> None:
        """
        Capture event close.
        """
        self.frameGeometry()
        if self.focusWidget():
            parent = self.focusWidget().parentWidget()  # type: ignore [union-attr]
            if parent:
                fdb = cast(flfielddb.FLFieldDB, parent)
                acf_ = getattr(fdb, "autoComFrame_", None)
                if acf_ and acf_.autoComFrame_.isVisible():
                    acf_.hide()
                    return

        if self.cursor_:
            if not self.cursor_.useDelegateCommit():
                try:
                    levels = self.cursor_.transactionLevel() - self._init_translation_level
                    if levels > 0:
                        self.cursor_.rollbackOpened(
                            levels,
                            "Se han detectado transacciones no finalizadas en la última operación.\n"
                            "Se van a cancelar las transacciones pendientes.\n"
                            "Los últimos datos introducidos no han sido guardados, por favor\n"
                            "revise sus últimas acciones y repita las operaciones que no\n"
                            "se han guardado.\n"
                            "FLFormRecordDB::closeEvent: %s %s" % (levels, self.name()),
                        )
                except Exception as error:
                    print(
                        "ERROR: FLFormRecordDB @ closeEvent :: las transacciones aún no funcionan.error: %s"
                        % (str(error))
                    )
            self.cursor_.restorePersistentFilterBeforeDelegate()

            if self.accepted_:
                if not self.cursor_.doCommit():
                    return
                if not self.cursor_.useDelegateCommit():
                    self.afterCommitTransaction()
            else:
                if not self.cursor_.useDelegateCommit():
                    if not self.cursor_.rollback():
                        event.ignore()  # type: ignore [union-attr]
                    else:
                        if not self.cursor().valueBuffer(self.cursor().primaryKey()):
                            self.cursor().refresh()

            self.setCursor(None)

        self.closed.emit()
        if event is not None:
            super().closeEvent(event)
        # self.deleteLater()

    def validateForm(self) -> bool:
        """
        Form validation.

        Call the "validateForm" function of the associated script when the
        form and only continue with the commit commit when that function
        of script returns TRUE.

        If FLTableMetaData :: concurWarn () is true and two or more sessions / users are.
        Modifying the same fields will display a warning notice.

        @return TRUE if the form has been validated correctly.
        """

        if not self.cursor_:
            return True
        # metadata = self.cursor_.metadata()

        # if self.cursor_.modeAccess() == pnsqlcursor.PNSqlCursor.Edit and metadata.concurWarn():
        #    col_fields = self.cursor_.concurrencyFields()

        #    if col_fields:
        #        pk_name = metadata.primaryKey()
        #        pk_where = (
        #            self.cursor_.db()
        #            .connManager()
        #            .manager()
        #            .formatAssignValue(metadata.field(pk_name), self.cursor_.valueBuffer(pk_name))
        #        )
        #        qry = pnsqlquery.PNSqlQuery(None, self.cursor_.db().connectionName())
        #        qry.setTablesList(metadata.name())
        #        qry.setSelect(col_fields)
        #        qry.setFrom(metadata.name())
        #        qry.setWhere(pk_where)
        #        qry.setForwardOnly(True)

        #        if qry.exec_() and qry.next():
        #            i = 0
        #            for field in col_fields:
        #                # msg = "El campo '%s' con valor '%s' ha sido modificado\npor otro usuario con el valor '%s'" % (
        #                #    mtd.fieldNameToAlias(field), self.cursor_.valueBuffer(field), q.value(i))
        #                res = QtWidgets.QMessageBox.warning(
        #                    QtWidgets.QApplication.focusWidget(),
        #                    "Aviso de concurrencia",
        #                    "\n\n ¿ Desea realmente modificar este campo ?\n\n"
        #                    "Sí : Ignora el cambio del otro usuario y utiliza el valor que acaba de introducir\n"
        #                    "No : Respeta el cambio del otro usuario e ignora el valor que ha introducido\n"
        #                    "Cancelar : Cancela el guardado del registro y vuelve a la edición del registro\n\n",
        #                    cast(
        #                        QtWidgets.QMessageBox.StandardButtons,
        #                        QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.Default,
        #                    ),
        #                    cast(
        #                        QtWidgets.QMessageBox.StandardButton,
        #                        QtWidgets.QMessageBox.No
        #                        | QtWidgets.QMessageBox.Cancel
        #                        | QtWidgets.QMessageBox.Escape,
        #                    ),
        #                )
        #                if res == QtWidgets.QMessageBox.Cancel:
        #                    return False

        #                if res == QtWidgets.QMessageBox.No:
        #                    self.cursor_.setValueBuffer(field, qry.value(i))

        if self.cursor_.modeAccess() in [self.cursor_.Insert, self.cursor_.Edit]:
            fun_ = getattr(self.iface, "validateForm", self.validateForm)
            if fun_ != self.validateForm:
                ret_ = fun_()

                if isinstance(ret_, bool):
                    return ret_

        return True

    def acceptedForm(self) -> None:
        """
        Accept of form.

        Call the "acceptedForm" function of the script associated with the form, when
        the form is accepted and just before committing the registration.
        """

        fun_ = getattr(self.iface, "acceptedForm", self.acceptedForm)
        if fun_ != self.acceptedForm:
            fun_()

    def afterCommitBuffer(self) -> None:
        """
        After setting the changes of the current record buffer.

        Call the "afterCommitBuffer" function of the script associated with the form
        right after committing the registry buffer.
        """

        fun_ = getattr(self.iface, "afterCommitBuffer", self.afterCommitBuffer)
        if fun_ != self.afterCommitBuffer:
            fun_()

    def afterCommitTransaction(self) -> None:
        """
        After fixing the transaction.

        Call the "afterCommitTransaction" function of the script associated with the form,
        right after finishing the current transaction accepting.
        """

        fun_ = getattr(self.iface, "afterCommitTransaction", self.afterCommitTransaction)
        if fun_ != self.afterCommitTransaction:
            fun_()

    def canceledForm(self) -> None:
        """
        Form Cancellation.

        Call the "canceledForm" function of the script associated with the form, when
        cancel the form.
        """
        fun_ = getattr(self.iface, "canceledForm", self.canceledForm)
        if fun_ != self.canceledForm:
            fun_()

    @decorators.pyqt_slot()
    def accept(self) -> None:
        """
        Activate pressing the accept button.
        """

        if self.accepting:
            return

        self.accepting = True

        if not self.cursor_:
            self.close()
            self.accepting = False
            return

        if not self.validateForm():
            self.accepting = False
            return

        if self.cursor_.checkIntegrity():
            self.acceptedForm()
            self.cursor_.setActivatedCheckIntegrity(False)
            if not self.cursor_.doCommitBuffer():
                self.accepting = False
                return
            else:
                self.cursor_.setActivatedCheckIntegrity(True)
        else:
            self.accepting = False
            return

        self.afterCommitBuffer()
        self.accepted_ = True
        self.close()
        self.accepting = False

    @decorators.pyqt_slot()
    def acceptContinue(self) -> None:
        """
        Activate pressing the accept and continue button.
        """
        if self.accepting:
            return

        self.accepting = True
        if not self.cursor_:
            self.close()
            self.accepting = False
            return

        if not self.validateForm():
            self.accepting = False
            return

        if self.cursor_.checkIntegrity():
            self.acceptedForm()
            self.cursor_.setActivatedCheckIntegrity(False)
            if self.cursor_.doCommitBuffer():
                self.cursor_.setActivatedCheckIntegrity(True)
                self.cursor_.doCommit()
                self.cursor_.setModeAccess(pnsqlcursor.PNSqlCursor.Insert)
                self.accepted_ = False
                caption = None
                if self._action:
                    caption = self._action.name()
                if not caption:
                    caption = self.cursor_.metadata().alias()
                if not self.cursor_.useDelegateCommit():
                    self.cursor_.transaction()
                self.setCaptionWidget(caption)
                if self._init_focus_widget:
                    self._init_focus_widget.setFocus()
                self.cursor_.refreshBuffer()
                self.initScript()

        self.accepting = False

    @decorators.pyqt_slot()
    def reject(self) -> None:
        """
        Activate pressing the cancel button.
        """
        self.accepted_ = False
        self.canceledForm()
        self.close()

    @decorators.pyqt_slot()
    def firstRecord(self) -> None:
        """
        Go to the first record.
        """
        if self.cursor_ and not self.cursor_.at() == 0:
            if not self.validateForm():
                return

            if self.cursor_.checkIntegrity():
                self.acceptedForm()
                self.cursor_.setActivatedCheckIntegrity(False)
                if self.cursor_.doCommitBuffer(False):
                    self.cursor_.setActivatedCheckIntegrity(True)
                    # self.cursor_.commit()
                    self.cursor_.setModeAccess(self._initial_mode_access)
                    self.accepted_ = False
                    # self.cursor_.transaction()
                    self.cursor_.first()
                    self.setCaptionWidget()
                    self.initScript()

    @decorators.pyqt_slot()
    def previousRecord(self) -> None:
        """
        Go to the previous record.
        """
        if self.cursor_ and self.cursor_.isValid():
            if not self.validateForm():
                return

            if self.cursor_.checkIntegrity():
                self.acceptedForm()
                self.cursor_.setActivatedCheckIntegrity(False)
                if self.cursor_.doCommitBuffer(False):
                    self.cursor_.setActivatedCheckIntegrity(True)
                    # self.cursor_.commit()
                    self.cursor_.setModeAccess(self._initial_mode_access)
                    self.accepted_ = False
                    # self.cursor_.transaction()
                    if self.cursor_.at() == 0:
                        self.cursor_.last()
                    else:
                        self.cursor_.prev()
                    self.setCaptionWidget()
                    self.initScript()

    @decorators.pyqt_slot()
    def nextRecord(self) -> None:
        """
        Go to the next record.
        """

        if self.cursor_ and self.cursor_.isValid():
            if not self.validateForm():
                return

            if self.cursor_.checkIntegrity():
                self.acceptedForm()
                self.cursor_.setActivatedCheckIntegrity(False)
                if self.cursor_.doCommitBuffer(False):
                    self.cursor_.setActivatedCheckIntegrity(True)
                    # self.cursor_.commit()
                    self.cursor_.setModeAccess(self._initial_mode_access)
                    self.accepted_ = False
                    # self.cursor_.transaction()
                    if self.cursor_.at() == (self.cursor_.size() - 1):
                        self.cursor_.first()
                    else:
                        self.cursor_.next()
                    self.setCaptionWidget()
                    self.initScript()

    @decorators.pyqt_slot()
    def lastRecord(self) -> None:
        """
        Go to the last record.
        """
        if (
            self.cursor_
            and not self.cursor_.at() == (self.cursor_.size() - 1)
            and self.cursor_.isValid()
        ):
            if not self.validateForm():
                return

            if self.cursor_.checkIntegrity():
                self.acceptedForm()
                self.cursor_.setActivatedCheckIntegrity(False)
                if self.cursor_.doCommitBuffer(False):
                    self.cursor_.setActivatedCheckIntegrity(True)
                    # self.cursor_.commit()
                    self.cursor_.setModeAccess(self._initial_mode_access)
                    self.accepted_ = False
                    # self.cursor_.transaction()
                    self.cursor_.last()
                    self.setCaptionWidget()
                    self.initScript()

    @decorators.pyqt_slot()
    def disablePushButtonCancel(self) -> None:
        """
        Turn off the cancel button.
        """

        if self.pushButtonCancel:
            self.pushButtonCancel.setDisabled(True)

    def show(self) -> None:
        """Show this widget."""

        caption = self._action.caption()
        if not caption:
            caption = self.cursor().metadata().alias()

        cur = self.cursor_

        iface = getattr(self.script, "iface", None)

        if cur:
            if not cur.isValid():
                cur.model().refresh()

            if cur.modeAccess() in (cur.Insert, cur.Edit, cur.Browse):
                if not cur.useDelegateCommit():
                    cur.transaction()
                    self._init_translation_level = cur.transactionLevel()

            if cur.modeAccess() == pnsqlcursor.PNSqlCursor.Insert:
                self._show_accept_continue = True
            else:
                self._show_accept_continue = False

            self.loadControls()

            if iface is not None:
                cur.setContext(iface)

        self.setCaptionWidget(caption)
        super().show()

    def inicializeControls(self) -> None:
        """Initialize UI controls for this form."""
        from pineboolib.fllegacy.flfielddb import FLFieldDB

        for child_ in self.findChildren(QtWidgets.QWidget):
            if isinstance(child_, FLFieldDB):
                loaded = getattr(child_, "_loaded", None)
                if loaded is False:
                    QtCore.QTimer.singleShot(0, child_.load)

    def show_and_wait(self) -> None:
        """Show this form blocking for exit."""
        if self.loop:
            raise Exception("show_and_wait(): Se ha detectado una llamada recursiva")

        self.loop = True
        self.show()
        if self.eventloop:
            self.eventloop.exec()

        self.loop = False

    def hide(self) -> None:
        """Hide this form."""
        if self.loop:
            self.eventloop.exit()



"""Qaction module."""

# -*- coding: utf-8 -*-
from PyQt6 import QtCore, QtGui  # type: ignore[import]
from typing import Optional


class QAction(QtGui.QAction):
    """QAction class."""

    activated = QtCore.pyqtSignal()
    _menu_text: str

    def __init__(self, *args) -> None:
        """Inicialize."""

        super().__init__(*args)
        self.triggered.connect(self.send_activated)  # type: ignore [attr-defined] # noqa: F821
        self._menu_text = ""

    def send_activated(self, value: Optional[bool] = None) -> None:
        """Send activated signal."""

        self.activated.emit()

    def getName(self) -> str:
        """Return widget name."""

        return self.objectName()

    def setName(self, name: str) -> None:
        """Set widget name."""

        self.setObjectName(name)

    def getMenuText(self) -> str:
        """Return menu text."""

        return self._menu_text

    def setMenuText(self, text: str) -> None:
        """Set menu text."""

        self._menu_text = text

    name = property(getName, setName)
    menuText = property(getMenuText, setMenuText)



"""
Unit Test Package for QSA parser.
"""
import os.path


def fixture_path(*path: str) -> str:
    """
    Get fixture path for this test folder.
    """
    basedir = os.path.realpath(os.path.dirname(__file__))
    filepath = os.path.join(basedir, "fixtures", *path)
    return filepath


def fixture_read(*path: str) -> str:
    """
    Read fixture from this test folder.
    """
    with open(fixture_path(*path), "r", encoding="utf-8") as file:
        contents = file.read()
    return contents



# -*- coding: utf-8 -*-
# Translated with pineboolib v0.71.18
"""Fltest4_model module."""

import sqlalchemy  # type: ignore [import] # noqa: F821

from pineboolib.application.database.orm import basemodel


class Fltest4(basemodel.BaseModel):  # type: ignore [misc] # noqa: F821
    """Fltest4 class."""

    __tablename__ = "fltest4"

    # --- Metadata --->
    legacy_metadata = {
        "name": "fltest4",
        "alias": "Test table",
        "fields": [
            {
                "name": "id",
                "alias": "ID",
                "pk": True,
                "type": "serial",
                "null": False,
                "visiblegrid": False,
                "editable": False,
            },
            {
                "name": "id_test",
                "alias": "Id. test",
                "type": "uint",
                "relations": [{"card": "M1", "table": "fltest", "field": "id"}],
            },
            {
                "name": "idarea",
                "alias": "Área",
                "type": "string",
                "length": 15,
                "relations": [{"card": "1M", "table": "fltest5", "field": "idarea"}],
                "null": False,
            },
            {
                "name": "other_field",
                "alias": "otro campo",
                "type": "string",
                "length": 15,
                "relations": [{"card": "1M", "table": "fltest1234567", "field": "id"}],
                "null": False,
            },
        ],
    }

    # <--- Metadata ---

    # --- Fields --->

    id = sqlalchemy.Column("id", sqlalchemy.Integer, primary_key=True)
    id_test = sqlalchemy.Column("id_test", sqlalchemy.Integer)
    idarea = sqlalchemy.Column("idarea", sqlalchemy.String(15))
    other_field = sqlalchemy.Column("other_field", sqlalchemy.String(15))


# <--- Fields ---



# -*- coding: utf-8 -*-
# Translated with pineboolib v0.71.18
"""Flmodules_model module."""

import sqlalchemy  # type: ignore [import] # noqa: F821
from pineboolib.application import TESTING_MODE
from pineboolib.application.database.orm import basemodel
from sqlalchemy.orm import relationship, foreign  # type: ignore [import]
from pineboolib.qsa import qsa
from typing import Optional


class Flmodules(basemodel.BaseModel):  # type: ignore [misc] # noqa: F821
    """Flmodules class."""

    __tablename__ = "flmodules"

    # --- Metadata --->
    legacy_metadata = {
        "name": "flmodules",
        "alias": "Módulos",
        "fields": [
            {
                "name": "bloqueo",
                "alias": "Bloqueo",
                "type": "unlock",
                "null": False,
                "default": True,
            },
            {
                "name": "idmodulo",
                "alias": "Id. del Módulo",
                "pk": True,
                "type": "string",
                "length": 15,
                "relations": [{"card": "1M", "table": "flfiles", "field": "idmodulo"}],
                "null": False,
            },
            {
                "name": "idarea",
                "alias": "Id. del Área",
                "type": "string",
                "length": 15,
                "relations": [{"card": "M1", "table": "flareas", "field": "idarea"}],
                "null": False,
                "visiblegrid": False,
            },
            {
                "name": "descripcion",
                "alias": "Descripción",
                "type": "string",
                "length": 100,
                "null": False,
            },
            {
                "name": "version",
                "alias": "Versión",
                "type": "string",
                "length": 3,
                "regexp": r"[0-9]\\.[0-9]",
                "null": False,
                "default": "0.0",
                "editable": False,
            },
            {"name": "icono", "alias": "Icono", "type": "pixmap"},
        ],
    }

    # <--- Metadata ---

    # --- Fields --->

    bloqueo = sqlalchemy.Column("bloqueo", sqlalchemy.Boolean)
    idmodulo = sqlalchemy.Column("idmodulo", sqlalchemy.String(15), primary_key=True)
    idarea = sqlalchemy.Column("idarea", sqlalchemy.String(15))
    descripcion = sqlalchemy.Column("descripcion", sqlalchemy.String(100))
    version = sqlalchemy.Column("version", sqlalchemy.String(3))
    icono = sqlalchemy.Column("icono", sqlalchemy.String)

    # <--- Fields ---

    @classmethod
    def pending_relationships(cls):
        """Pending relationships."""
        if not TESTING_MODE:
            return True

        areas_class = qsa.orm_("flareas", False)
        if areas_class and not hasattr(areas_class, "children"):
            areas_class.children = relationship(
                cls,
                primaryjoin=areas_class.idarea == foreign(cls.idarea),
                cascade="save-update, merge, delete, delete-orphan",
            )
            return True

        return False

    def before_flush(self) -> Optional[bool]:
        """Before flush."""
        if TESTING_MODE:
            if self.idmodulo == "T2M1_1":
                raise Exception("before_flush_called!")

        return True



# -*- coding: utf-8 -*-
"""
PNAccessControlFactory Module.

Manage ACLs between different application objects.
"""

from PyQt6 import QtWidgets, QtGui  # type: ignore[import]

from pineboolib.application.metadata import pntablemetadata
from pineboolib.application.acls import pnaccesscontrol


from typing import Dict, Union, cast, List

from pineboolib import logging

LOGGER: "logging.Logger" = logging.get_logger(__name__)


class PNAccessControlMainWindow(pnaccesscontrol.PNAccessControl):
    """PNAccessControlMainWindow Class."""

    def type(self) -> str:
        """Return target type."""

        return "mainwindow"

    def processObject(  # type: ignore [override]
        self, main_window: "QtWidgets.QMainWindow"
    ) -> None:
        """Process the object."""

        if self._perm:
            for action in main_window.findChildren(QtGui.QAction):
                action_name = action.objectName()
                perms_value = (
                    self._acos_perms[action_name]
                    if action_name in self._acos_perms.keys()
                    else self._perm
                )
                if perms_value in ("-w", "--"):
                    cast(QtWidgets.QWidget, action).setVisible(False)


class PNAccessControlForm(pnaccesscontrol.PNAccessControl):
    """PNAccessControlForm Class."""

    def __init__(self) -> None:
        """Inicialize."""

        super().__init__()

        self.pal = QtGui.QPalette()
        palette_ = QtWidgets.QApplication.palette()  # type: ignore[misc] # noqa: F821
        background_color = palette_.color(
            QtGui.QPalette.ColorGroup.Active, QtGui.QPalette.ColorRole.Base
        )
        # FIXMEPYQT6
        # self.pal.setColor(QtGui.QPalette.Foreground, background_color)
        self.pal.setColor(QtGui.QPalette.ColorRole.Text, background_color)
        self.pal.setColor(QtGui.QPalette.ColorRole.ButtonText, background_color)
        self.pal.setColor(QtGui.QPalette.ColorRole.Base, background_color)
        # self.pal.setColor(QtGui.QPalette.Background, background_color)

    def type(self) -> str:
        """Return target type."""
        return "form"

    def processObject(self, widget: "QtWidgets.QWidget") -> None:  # type: ignore [override]
        """
        Process objects that are of the FLFormDB class.

        Only control the children of the object that are of the QWidget class, and only
        allows to make them not visible or not editable. Actually do them
        not visible means that they are not editable and modifying the palette to
        that the entire region of the component be shown in black. The permits
        which accepts are:

        - "-w" or "--" (no_read / write or no_read / no_write) -> not visible
        - "r-" (read / no_write) -> not editable

        This allows any component of an AbanQ form (FLFormDB,
        FLFormRecordDB and FLFormSearchDB) can be made not visible or not editable for convenience.
        """

        not_found_widgets: List[str] = list(self._acos_perms.keys())

        for child_object in widget.findChildren(QtWidgets.QWidget):
            child_widget = cast(QtWidgets.QWidget, child_object)
            widget_name = child_widget.objectName()
            perms = self._perm
            if widget_name in self._acos_perms.keys():
                perms = self._acos_perms[widget_name]
                not_found_widgets.remove(widget_name)

            if perms == "":
                continue

            elif perms in ("-w", "--"):
                child_widget.setPalette(self.pal)
                child_widget.setDisabled(True)
                child_widget.hide()

            elif perms == "r-":
                child_widget.setDisabled(True)

        for not_found_widget_name in not_found_widgets:
            LOGGER.warning(
                "PNAccessControlFactory: No se encuentra el control %s para procesar ACLS.",
                not_found_widget_name,
            )


class PNAccessControlTable(pnaccesscontrol.PNAccessControl):
    """PNAccessControlTable Class."""

    def __init__(self) -> None:
        """Inicialize."""

        super().__init__()
        self._acos_perms: Dict[str, str] = {}

    def type(self) -> str:
        """Return target type."""

        return "table"

    def processObject(  # type: ignore [override]
        self, table_metadata: "pntablemetadata.PNTableMetaData"
    ) -> None:
        """Process pntablemetadata.PNTableMetaData belonging to a table."""
        mask_perm = 0
        has_acos = True if self._acos_perms else False

        if self._perm:
            if self._perm[0] == "r":
                mask_perm += 2
            if self._perm[1] == "w":
                mask_perm += 1
        elif has_acos:
            mask_perm = 3
        else:
            return

        for field in table_metadata.fieldList():
            mask_field_perm = mask_perm  # por defecto valores de self._perm

            if has_acos and (
                field.name() in self._acos_perms.keys()
            ):  # si hay acos_perm especifico sobrecargo
                field_perm = self._acos_perms[field.name()]
                mask_field_perm = 0
                if field_perm[0] == "r":
                    mask_field_perm += 2

                if field_perm[1] == "w":
                    mask_field_perm += 1

            if mask_field_perm == 0:
                field.setVisible(False)
                field.setEditable(False)
            elif mask_field_perm == 1:
                if not field.visible():
                    continue
                else:
                    field.setVisible(True)
                field.setEditable(False)
            elif mask_field_perm == 2:
                field.setVisible(True)
                field.setEditable(False)
            elif mask_field_perm == 3:
                field.setVisible(True)
                field.setEditable(True)

    def setFromObject(self, table_mtd: "pntablemetadata.PNTableMetaData") -> None:
        """Apply permissions from a pntablemetadata.PNTableMetaData."""

        self._acos_perms.clear()

        for field in table_mtd.fieldList():
            perm_read = "r" if field.visible() else "-"
            perm_write = "w" if field.editable() else "-"
            self._acos_perms[field.name()] = "%s%s" % (perm_read, perm_write)


class PNAccessControlFactory(object):
    """PNAccessControlFactory Class."""

    def create(self, type_: str = "") -> "pnaccesscontrol.PNAccessControl":
        """Create a control instance according to the type that we pass."""

        if not type_:
            raise ValueError("type_ must be set")

        if type_ == "mainwindow":
            return PNAccessControlMainWindow()
        elif type_ == "form":
            return PNAccessControlForm()
        elif type_ == "table":
            return PNAccessControlTable()

        raise ValueError("type_ %r unknown" % type_)

    def type(
        self,
        obj: Union["QtWidgets.QWidget", "pntablemetadata.PNTableMetaData", "QtWidgets.QMainWindow"],
    ) -> str:
        """Return the type of instance target."""

        ret_: str = ""

        if isinstance(obj, QtWidgets.QMainWindow):
            ret_ = "mainwindow"
        elif isinstance(obj, pntablemetadata.PNTableMetaData):
            ret_ = "table"
        elif isinstance(obj, QtWidgets.QDialog):
            ret_ = "form"

        return ret_



"""
Data Types for QSA.
"""

import codecs

import os
import collections


from PyQt6 import QtCore  # type: ignore [import]

from pineboolib.core.utils import utils_base
from pineboolib.core import decorators

from pineboolib.application.qsatypes.date import Date  # noqa: F401
from pineboolib.application.utils import modules
from pineboolib import logging

from typing import Any, Optional, Dict, Union, Generator, List


LOGGER = logging.get_logger(__name__)


def boolean(value: Union[bool, str, float] = False) -> bool:
    """
    Return boolean from string.
    """
    if isinstance(value, bool):
        return value
    elif isinstance(value, str):
        value = value.lower().strip()[0]
        if value in ["y", "t"]:
            return True
        if value in ["n", "f"]:
            return False
        raise ValueError("Cannot convert %r to Boolean" % value)
    elif isinstance(value, int):
        return value != 0
    elif isinstance(value, float):
        if abs(value) < 0.01:
            return False
        else:
            return True
    else:
        raise ValueError("Cannot convert %r to Boolean" % value)


class QString(str):
    """
    Emulate original QString as was removed from PyQt6.
    """

    def mid(self, start: int, length: int = 0) -> str:
        """
        Cut sub-string.

        @param start. Posición inicial
        @param length. Longitud de la cadena. Si no se especifica , es hasta el final
        @return sub cadena de texto.
        """

        return self[start:] if not length else self[start : start + length]

    @staticmethod
    def fromCharCode(*args: int) -> str:
        """Return a char list values."""

        return "".join([chr(val) for val in args])


def function(*args: str) -> Any:
    """
    Load QS string code and create a function from it.

    Parses it to Python and return the pointer to the function.
    """

    # Leer código QS embebido en Source
    # asumir que es una funcion anónima, tal que:
    #  -> function($args) { source }
    # compilar la funcion y devolver el puntero
    source_pos = len(args) - 1
    arguments = args[:source_pos]
    source = args[source_pos]
    qs_source = """

function anon(%s) {
    %s
} """ % (
        ", ".join(arguments),
        source,
    )

    module = modules.text_to_module(qs_source)

    return module.anon


def object_(value: Optional[Dict[str, Any]] = None) -> "utils_base.StructMyDict":
    """
    Object type "object".
    """
    return utils_base.StructMyDict(value or {})


String = QString


class Array(object):
    """
    Array type object.
    """

    # NOTE: To avoid infinite recursion on getattr/setattr, all attributes MUST be defined at class-level.
    _dict: Dict[Any, Any] = {}
    _pos_iter = 0

    def __init__(self, *args: Any) -> None:
        """Create new array."""
        self._pos_iter = 0
        self._dict = collections.OrderedDict()

        if not len(args):
            return
        elif len(args) == 1:
            if isinstance(args[0], list):
                for key, value in enumerate(args[0]):
                    self._dict[key] = value

            elif isinstance(args[0], dict):
                for key, value in args[0].items():
                    self._dict[key] = value

            elif isinstance(args[0], int):
                return

        elif isinstance(args[0], str):
            for item in args:
                self.__setitem__(item, item)

    def __iter__(self) -> Generator[Any, None, None]:
        """
        Iterate through values.
        """
        for value in list(self._dict.values()):
            yield value

    def __setitem__(self, key: Union[str, int], value: Any) -> None:
        """
        Set item.

        @param key. Nombre del registro
        @param value. Valor del registro
        """
        # field_key = key
        # while field_key in self.dict_.keys():
        #    field_key = "%s_bis" % field_key
        self._dict[key] = value

    def __getitem__(self, key: Union[str, int, slice]) -> Any:
        """
        Get item.

        @param key. Valor que idenfica el registro a recoger
        @return Valor del registro especificado
        """
        result = None
        if isinstance(key, int):
            try:
                result = list(self._dict.values())[key]
            except Exception:
                pass

        elif isinstance(key, slice):
            LOGGER.warning("FIXME: Array __getitem__%s con slice" % key)
        else:
            try:
                result = self._dict[key]
            except Exception:
                pass

        return result

    def length(self) -> int:
        """Return array size."""
        return len(self._dict)

    def __getattr__(self, value: str) -> Any:
        """Support for attribute style access."""
        return self._dict[value] if value in self._dict.keys() else None

    def __setattr__(self, name: str, value: Any) -> None:
        """Support for attribute style writes."""
        if name[0] == "_":
            return super().__setattr__(name, value)
        self._dict[name] = value

    def __eq__(self, other: Any) -> bool:
        """Support for equality comparisons."""
        if isinstance(other, Array):
            return other._dict == self._dict
        elif isinstance(other, list):
            return other == list(self._dict.values())
        elif isinstance(other, dict):
            return other == self._dict
        return False

    def __repr__(self) -> str:
        """Support for repr."""
        return "<%s %r>" % (self.__class__.__name__, list(self._dict.values()))

    def splice(self, *args: Any) -> None:
        """Cut or replace array."""
        new_dict = {}

        if len(args) == 2:  # Delete
            for key in list(self._dict.keys())[args[0] : args[0] + args[1]]:
                new_dict[key] = self._dict[key]

        elif len(args) > 2 and args[1] == 0:  # Insertion
            fix_pos = 0
            for pos in range(len(self._dict.keys())):
                new_dict[pos + fix_pos] = self._dict[pos]
                if pos == args[0]:
                    for new_value in args[2:]:
                        fix_pos += 1
                        new_dict[pos + fix_pos] = new_value

        elif len(args) > 2 and args[1] > 0:  # Replacement
            for pos, key in enumerate(self._dict.keys()):
                if pos == args[0]:
                    for new_value in args[2:]:
                        new_dict[new_value] = new_value
                else:
                    if pos < args[0] or pos >= args[0] + args[1]:
                        new_dict[key] = self._dict[key]

        self._dict = new_dict

    def __len__(self) -> int:
        """Return size of array."""
        return len(self._dict.keys())

    def __str__(self) -> str:
        """Support for str."""
        return repr(list(self._dict.values()))

    def append(self, value: Any) -> None:
        """Append new value."""
        size = len(self._dict.keys())
        while size in self._dict:
            size += 1
        self._dict[size] = value

    def shift(self) -> Any:
        """Shifts (i.e. removes) the bottom-most (left-most) item off the array and returns it."""
        return self.pop(0)

    def pop(self, position: int = 0) -> Any:
        """Return a position value and delte it from list."""

        value = self._dict[position]
        self._dict[position] = None
        del self._dict[position]

        return value

    def concat(
        self, *args: Union[List[Any], Dict[str, Any], "Array"]
    ) -> Union[List[Any], Dict[str, Any]]:
        """Return arrays concatenated."""

        if len(args) > 1 and isinstance(args[1], (list, Array)):
            result_list: List[Any] = []
            for item in args:
                result_list += item

            return result_list
        elif len(args) == 1 and isinstance(args[0], (list, Array)):
            for item in args[0]:
                self.append(item)
            return self  # type: ignore [return-value]

        else:
            result_array: Dict[str, Any] = {}
            for item in args:
                for key, value in item.items():  # type: ignore [union-attr]
                    result_array[key] = value

            return result_array


AttributeDict = utils_base.StructMyDict


class Dir(object):
    """
    Manage folder.

    Emulates QtCore.QDir for QSA.
    """

    path: Optional[str]

    # Filters :
    Files = QtCore.QDir.Filter.Files
    Dirs = QtCore.QDir.Filter.Dirs
    NoFilter = QtCore.QDir.Filter.NoFilter

    # Sort Flags:
    Name = QtCore.QDir.SortFlag.Name
    NoSort = QtCore.QDir.SortFlag.NoSort

    # other:
    home = os.path.expanduser("~")

    def __init__(self, path: Optional[str] = None):
        """Create a new Dir."""
        self.path = path

    def entryList(
        self,
        patron: str,
        filter: "QtCore.QDir.Filter" = QtCore.QDir.Filter.NoFilter,
        sort: "QtCore.QDir.SortFlag" = QtCore.QDir.SortFlag.NoSort,
    ) -> list:
        """
        Create listing for files inside given folder.

        @param patron. Patron a usa para identificar los ficheros
        @return lista con los ficheros que coinciden con el patrón
        """
        return QtCore.QDir(self.path).entryList([patron], filter, sort)  # type: ignore [arg-type]

    @staticmethod
    def fileExists(file_name: str) -> bool:
        """
        Check if a file does exist.

        @param file_name. Nombre del fichero
        @return Boolean. Si existe el ficehro o no.
        """
        return os.path.exists(file_name)

    @staticmethod
    def cleanDirPath(name: str) -> str:
        """
        Clean path from unnecesary folders.
        """
        return os.path.normpath(name)

    @staticmethod
    @decorators.deprecated
    def convertSeparators(filename: str) -> str:
        """
        Convert path from backslash to slashes or viceversa.

        ***DEPRECATED***
        """
        return filename

    @staticmethod
    def setCurrent(val: Optional[str] = None) -> None:
        """
        Change current working folder.

        @param val. Ruta especificada
        """
        os.chdir(val or utils_base.filedir("."))

    def getCurrent(self) -> str:
        """Return current folder."""
        return os.curdir

    def set_current(self, new_path: Optional[str] = None) -> None:
        """Set new patch."""
        os.chdir(new_path or utils_base.filedir("."))

    def mkdir(self, name: str = "") -> None:
        """
        Create a new folder.

        @param name. Nombre de la ruta a crear
        """
        if not name and self.path is None:
            raise ValueError("self.path is not defined!")

        if self.path:
            name = self.path + "/" + name

        try:
            os.stat(name)
        except Exception:
            os.mkdir(name)

    def cd(self, path: str) -> None:
        """Change dir."""

        os.chdir(path)

    def cdUp(self) -> None:
        """Change directory by moving one directory up from the Dir's current directory if possible."""

        os.chdir("..")

    def rmdirs(self, name: Optional[str] = None) -> None:
        """Delete a folder."""

        if name is None:
            raise ValueError("name is not defined!")

        if self.path is None:
            raise ValueError("self.path is not defined!")

        path_ = os.path.join(self.path, name)

        if os.path.exists(path_):
            import shutil

            shutil.rmtree(path_)

    current = property(getCurrent, set_current)


DirStatic = Dir()


class FileBaseClass(object):
    """
    Constants for File and FileStatic.
    """

    ReadOnly = QtCore.QIODevice.OpenModeFlag.ReadOnly
    WriteOnly = QtCore.QIODevice.OpenModeFlag.WriteOnly
    ReadWrite = QtCore.QIODevice.OpenModeFlag.ReadWrite
    Append = QtCore.QIODevice.OpenModeFlag.Append
    ioDevice = QtCore.QIODevice

    @staticmethod
    def exists(name: str) -> bool:
        """
        Check if a file does exist.

        @param name. Nombre del fichero.
        @return boolean informando si existe o no el fichero.
        """
        return os.path.exists(name)

    @staticmethod
    def isDir(dir_name: str) -> bool:
        """
        Check if given path is a folder.

        @param. Nombre del directorio
        @return. boolean informando si la ruta dada es un directorio o no.
        """
        return os.path.isdir(dir_name)

    @staticmethod
    def isFile(file_name: str) -> bool:
        """
        Check if given path is a file.

        @param. Nombre del fichero
        @return. boolean informando si la ruta dada es un fichero o no.
        """
        return os.path.isfile(file_name)


class File(FileBaseClass):  # FIXME : Rehacer!!
    """
    Manage a file.
    """

    _file_name: str
    _mode: "QtCore.QIODevice.OpenModeFlag"

    _encode: str
    _last_seek: int
    _q_file: QtCore.QFile
    eof: bool

    def __init__(self, file_path: Optional[str] = None, encode: Optional[str] = None):
        """Create a new File Object. This does not create a file on disk."""

        self._encode = "iso-8859-15"
        self._last_seek = 0
        self._file_name = ""
        self.eof = False

        if file_path is not None:
            self._q_file = QtCore.QFile(file_path)
            file_name, extension = os.path.splitext(file_path)
            self._file_name = "%s%s" % (file_name, extension)

        if encode is not None:
            self._encode = encode

        self._mode = self.ReadWrite

    def open(self, mode: "QtCore.QIODevice.OpenModeFlag") -> bool:
        """Open file."""

        self._mode = mode
        self.eof = False
        if self._q_file is not None:
            self._q_file.open(self._mode)

        return True

    def ioDevice(self) -> "QtCore.QIODevice":  # type: ignore [override] # noqa: F821
        """Return ioDevice mode."""
        return self._q_file

    def close(self) -> None:
        """Close file."""
        if self._q_file is not None:
            self._q_file.close()

    def errorString(self) -> str:
        """Return error string."""
        return self._q_file.errorString()

    def read(self, bytes_: bool = False) -> str:
        """
        Read file completely.

        @param bytes. Especifica si se lee en modo texto o en bytes
        @retunr contenido del fichero
        """

        if not self._file_name:
            raise ValueError("self._file_name is not defined!")
        if not bytes_:
            file_ = codecs.open(self._file_name, "r", encoding=self._encode)
        else:
            file_ = open(self._file_name, "rb")  # type: ignore [assignment] # noqa: F821
        ret = file_.read()
        file_.close()
        self.eof = True
        return ret

    def readAll(self) -> str:
        """Read file completely."""
        return self.read(True)

    def write(self, data: Union[str, bytes], length: int = -1) -> None:
        """
        Write data back to the file.

        @param data. Valores a guardar en el fichero
        @param length. Tamaño de data. (No se usa)
        """
        if not self._file_name:
            raise ValueError("self._file_name is empty!")

        mode = "a" if self._mode == self.Append else "w"

        if not isinstance(data, str):
            data = data.decode(self._encode)

        file_ = codecs.open(self._file_name, mode, encoding=self._encode, errors="replace")
        file_.write(data)
        file_.close()

    def writeBlock(self, data: bytes) -> None:
        """Write a block of data to the file."""
        if not self._file_name:
            raise ValueError("self._file_name is empty!")

        data_string = data.decode(self._encode)

        file_ = codecs.open(self._file_name, "w", encoding=self._encode, errors="replace")
        file_.write(data_string)
        file_.close()

    def getName(self) -> str:
        """
        Get file name.

        @return Nombre del _file_name
        """
        if not self._file_name:
            raise ValueError("self._file_name is empty!")

        _path, file_name = os.path.split(self._file_name)
        return file_name

    def getBaseName(self) -> str:
        """Return baseName."""

        if not self._file_name:
            raise ValueError("self._file_name is empty!")

        return os.path.basename(self._file_name.split(".")[0])

    def getExtension(self) -> str:
        """Return extension."""
        if not self._file_name:
            raise ValueError("self._file_name is empty!")

        return os.path.splitext(self._file_name)[1]

    def writeLine(self, data: str, len: Optional[int] = None) -> None:
        """
        Write a new line with "data" contents into the file.

        @param data. Datos a añadir en el _file_name
        """
        import codecs

        if not self._file_name:
            raise ValueError("self._file_name is empty!")

        file_ = codecs.open(self._file_name, encoding=self._encode, mode="a")
        file_.write("%s\n" % data if len is None else data[0:len])
        file_.close()

    def readLine(self) -> str:
        """
        Read a line from file.

        @return cadena de texto con los datos de la linea actual
        """

        import codecs

        if not self._file_name:
            raise ValueError("self._file_name is empty!")

        file_ = codecs.open(self._file_name, "r", encoding=self._encode)
        file_.seek(self._last_seek)
        ret = file_.readline(self._last_seek)
        self._last_seek += len(ret)
        self.eof = True if ret else False

        file_.close()

        return ret

    def readLines(self) -> List[str]:
        """
        Read all lines from a file and return it as array.

        @return array con las lineas del _file_name.
        """

        if not self._file_name:
            raise ValueError("self._file_name is empty!")

        ret: List[str]
        import codecs

        file_ = codecs.open(self._file_name, encoding=self._encode, mode="a")
        file_.seek(self._last_seek)
        ret = file_.readlines()
        file_.close()
        return ret

    def readbytes(self) -> bytes:
        """
        Read the whole file as binary.

        @return bytess con los datos de la linea actual
        """
        ret_ = self.read(True)

        return ret_.encode(self._encode)

    def writebytes(self, data_b: bytes) -> None:
        """
        Write file as binary.

        @param data_b. Datos a añadir en el _file_name
        """
        if not self._file_name:
            raise ValueError("self._file_name is empty!")

        file_ = open(self._file_name, "wb")
        file_.write(data_b)
        file_.close()

    def readByte(self) -> bytes:
        """Read a byte from file."""
        if not self.eof:
            with open(self._file_name, "rb") as file_:
                file_.seek(self._last_seek)
                self._last_seek += 1
                ret = file_.read(1)
                self.eof = True if not ret else False
                return ret

        return b""

    def writeByte(self, data: bytes) -> None:
        """Write a byte to file."""
        with open(self._file_name, "wb") as file_:
            file_.seek(self._last_seek)
            self._last_seek += 1
            file_.write(data)

    def remove(self) -> bool:
        """
        Delete file from filesystem.

        @return Boolean . True si se ha borrado el _file_name, si no False.
        """
        return self._q_file.remove()

    def getFullName(self) -> str:
        """Return full name."""
        return self._file_name or ""

    def getSize(self) -> int:
        """Return file size."""

        return self._q_file.size()

    def getPath(self) -> str:
        """Return getPath."""

        return os.path.abspath(os.path.dirname(self._file_name))

    def readable(self) -> bool:
        """Return if file is readable."""

        return os.access(self._file_name or "", os.R_OK)

    def lastModified(self) -> str:
        """Return last modified date."""

        return QtCore.QFileInfo(self._q_file).lastModified().toString("yyyy-MM-dd-hh:mm:ss")

    def exists(self) -> bool:  # type: ignore [override] # noqa: F821
        """Return True if exists a file else False."""

        return self._q_file.exists()

    name = property(getName)
    path = property(getPath)
    fullName = property(getFullName)
    baseName = property(getBaseName)
    extension = property(getExtension)
    size = property(getSize)


class FileStatic(FileBaseClass):
    """
    Static methods for File that overlap in name.
    """

    @staticmethod
    def remove(file_name: str) -> bool:
        """
        Delete file from filesystem.

        @return Boolean . True si se ha borrado el fichero, si no False.
        """
        file = File(file_name)
        return file.remove()

    @staticmethod
    def read(file_: str, bytes: bool = False) -> Union[str, bytes]:
        """
        Read file completely.

        @param bytes. Especifica si se lee en modo texto o en bytes
        @return contenido del fichero
        """

        with codecs.open(file_, "r" if not bytes else "rb", encoding="ISO-8859-15") as file_obj:
            ret = file_obj.read()

        return ret

    @staticmethod
    def write(file_: str, data: Union[str, bytes], length: int = -1) -> None:
        """
        Write data back to the file.

        @param data. Valores a guardar en el fichero
        @param length. Tamaño de data. (No se usa)
        """

        bytes_ = data.encode("ISO-8859-15") if isinstance(data, str) else data

        with open(file_, "wb") as file:
            file.write(bytes_)

        file.close()



from pineboolib import application, logging
import os
import importlib

from typing import Callable

LOGGER = logging.get_logger(__name__)


def load_project_config_file() -> None:
    """Load project config."""
    if application.EXTERNAL_FOLDER and application.PROJECT_NAME:

        path_config = os.path.abspath(
            os.path.join(application.EXTERNAL_FOLDER, "apps", application.PROJECT_NAME, "config.py")
        )
        LOGGER.info("PROJECT_NAME: %s, CONFIG: %s" % (application.PROJECT_NAME, path_config))
        if os.path.exists(path_config):
            from pineboolib.application.load_script import import_path

            mod_ = import_path("config_project", path_config)
            launch_function(mod_, "cargar_dependencias")
        else:
            LOGGER.warning("Config file not found: %s", path_config)


def reload_project_config() -> None:
    """Reload project config."""
    if application.EXTERNAL_FOLDER and application.PROJECT_NAME:
        LOGGER.warning("STATIC LOADER: Reinitializing project config file...")
        module_name = "apps.%s.config" % (application.PROJECT_NAME)
        try:
            mod_ = importlib.import_module(module_name)
            launch_function(mod_, "cargar_dependencias")

        except Exception as error:
            LOGGER.warning(
                "STATIC LOADER: Error reloading project config file %s, Error: %s"
                % (module_name, str(error))
            )


def launch_function(mod_: "Callable", func_name: str = None) -> None:
    """Launch function."""

    func_ = getattr(mod_, func_name, None)
    if func_:
        LOGGER.info("EXTERNAL: %s function found in %s" % (func_name, mod_.__name__))
        func_()
    else:
        LOGGER.warning("STATIC LOADER: No %s function found in %s" % (func_name, mod_.__name__))



"""Test Signals module."""

import unittest

from pineboolib.loader.main import init_testing, finish_testing
from pineboolib import application
from pineboolib.qsa import qsa

VALUE_1: int = 0
VALUE_2: str = ""
VALUE_3: bool = False
VALUE_4: str = ""
VALUE_5 = None


def update_value(field_name: str) -> None:
    """Update test value 1."""

    global VALUE_1, VALUE_2

    VALUE_2 = field_name
    VALUE_1 += 1


def update_value_2() -> None:
    """Update test value 2."""

    global VALUE_1

    VALUE_1 += 1


def update_value_3(field_name: str = "", cursor=None) -> None:
    """Update test value 3."""

    global VALUE_3

    VALUE_3 = field_name != "" and cursor is not None


def update_value_4(field_name: str = "", cursor=None) -> None:
    """Update test value 4."""

    global VALUE_4, VALUE_5

    if cursor.valueBuffer("idarea") == "JUAS_3":
        return

    VALUE_4 = cursor.valueBuffer(field_name)
    cursor.setValueBuffer("idarea", "JUAS_3")
    VALUE_5 = cursor


class TestSignals(unittest.TestCase):
    """TestQueryOrm Class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()
        application.PROJECT.conn_manager.manager().createTable("fltest4")
        application.PROJECT.conn_manager.manager().createTable("fltest5")

    def test_basic_1(self) -> None:
        global VALUE_1

        VALUE_1 = 0
        qsa.thread_session_new()
        obj_ = qsa.orm.fltest4()
        self.assertTrue(obj_)
        obj_.bufferChanged.connect(update_value)
        obj_.idarea = "juas"
        qsa.thread_session_free()
        self.assertEqual(VALUE_1, 1)

    def test_basic_2(self) -> None:
        global VALUE_1

        VALUE_1 = 0
        qsa.thread_session_new()
        obj_ = qsa.orm.fltest4()
        self.assertTrue(obj_)
        obj_.bufferChanged.connect(update_value)
        obj_.idarea = "juas"
        obj_.other_field = "dos"

        self.assertEqual(VALUE_1, 2)
        obj_.bufferChanged.disconnect(update_value)
        obj_.idarea = "juas2"
        self.assertEqual(VALUE_1, 2)
        qsa.thread_session_free()

    def test_basic_3(self) -> None:
        global VALUE_2

        VALUE_2 = ""
        qsa.thread_session_new()
        obj_ = qsa.orm.fltest4()
        self.assertTrue(obj_)
        obj_.bufferChanged.connect(update_value)
        obj_.idarea = "juas"
        qsa.thread_session_free()
        self.assertEqual(VALUE_2, "idarea")

    def test_basic_4(self) -> None:
        global VALUE_1, VALUE_2

        VALUE_1 = 0
        VALUE_2 = ""
        qsa.thread_session_new()
        obj_ = qsa.orm.fltest4()
        self.assertTrue(obj_)
        obj_.bufferChanged.connect(update_value)
        obj_.bufferChanged.connect(update_value_2)
        obj_.idarea = "juas"
        qsa.thread_session_free()
        self.assertEqual(VALUE_1, 2)
        self.assertEqual(VALUE_2, "idarea")

    def test_basic_5(self) -> None:
        global VALUE_1, VALUE_2

        VALUE_1 = 0
        VALUE_2 = ""
        qsa.thread_session_new()
        obj_ = qsa.orm.fltest4()
        self.assertTrue(obj_)
        obj_.cursor.bufferChanged.connect(update_value)
        obj_.bufferChanged.connect(update_value_2)
        self.assertEqual(obj_.cursor.bufferChanged, obj_.bufferChanged)
        obj_.idarea = "juas"
        qsa.thread_session_free()
        self.assertEqual(VALUE_1, 2)
        self.assertEqual(VALUE_2, "idarea")

    def test_basic_6(self) -> None:
        global VALUE_3

        qsa.thread_session_new()
        obj_ = qsa.orm.fltest4()
        self.assertTrue(obj_)
        obj_.cursor.bufferChanged.connect(update_value_3)
        obj_.idarea = "juas"

        self.assertEqual(VALUE_3, True)

    def test_basic_7(self) -> None:
        global VALUE_4, VALUE_5

        qsa.thread_session_new()
        obj_ = qsa.orm.fltest4()
        self.assertTrue(obj_)
        obj_.cursor.bufferChanged.connect(update_value_4)
        obj_.idarea = "juas"
        self.assertEqual(VALUE_4, "juas")
        self.assertTrue(VALUE_5)
        if VALUE_5 is not None:
            VALUE_5.setValueBuffer("idarea", "JUAS_2")
            self.assertEqual(obj_.idarea, "JUAS_2")
            VALUE_5.setValueBuffer("idarea", "JUAS_4")
            VALUE_5.setValueBuffer("idarea", "JUAS_3")
        self.assertEqual(obj_.idarea, "JUAS_3")
        self.assertEqual(obj_.changes(), {"id": 7, "idarea": "JUAS_3"})
        qsa.thread_session_free()
        self.assertEqual(VALUE_4, "JUAS_4")
        self.assertEqual(obj_.cursor.valueBuffer("idarea"), "JUAS_3")
        qsa.thread_session_free()

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""

        finish_testing()



# -*- coding: utf-8 -*-
# Translated with pineboolib v0.71.18
"""Flfiles_model module."""

from pineboolib.application.database.orm import basemodel

import sqlalchemy  # type: ignore [import]

from pineboolib.qsa import qsa

from typing import Optional


class Flfiles(basemodel.BaseModel):  # type: ignore [misc] # noqa: F821
    """Flfiles class."""

    __tablename__ = "flfiles"

    # --- Metadata --->
    legacy_metadata = {
        "name": "flfiles",
        "alias": "Ficheros de Texto",
        "fields": [
            {
                "name": "nombre",
                "alias": "Nombre",
                "pk": True,
                "type": "string",
                "length": 255,
                "null": False,
            },
            {
                "name": "bloqueo",
                "alias": "Bloqueo",
                "type": "unlock",
                "null": False,
                "default": True,
            },
            {
                "name": "idmodulo",
                "alias": "Módulo",
                "type": "string",
                "length": 15,
                "relations": [
                    {"card": "M1", "table": "flmodules", "field": "idmodulo", "delC": True}
                ],
                "null": False,
            },
            {
                "name": "sha",
                "alias": "SHA1",
                "type": "string",
                "length": 255,
                "calculated": True,
                "editable": False,
            },
            {"name": "contenido", "alias": "Contenido", "type": "stringlist", "visiblegrid": False},
            {"name": "binario", "alias": "Binario", "type": "bytearray", "visiblegrid": False},
        ],
    }

    # <--- Metadata ---

    # --- Fields --->

    nombre = sqlalchemy.Column("nombre", sqlalchemy.String(255), primary_key=True)
    bloqueo = sqlalchemy.Column("bloqueo", sqlalchemy.Boolean)
    idmodulo = sqlalchemy.Column("idmodulo", sqlalchemy.String(15))
    sha = sqlalchemy.Column("sha", sqlalchemy.String(255))
    contenido = sqlalchemy.Column("contenido", sqlalchemy.String)
    binario = sqlalchemy.Column("binario", sqlalchemy.LargeBinary)

    # <--- Fields ---

    def before_flush(self) -> Optional[bool]:
        """Before flush."""

        pass

    def after_flush(self) -> bool:
        """After flush."""

        session = self.session
        flfiles_class = qsa.from_project("flfiles_orm")
        flserial_class = qsa.from_project("flserial_orm")

        value = str(self.sha)
        util = qsa.FLUtil()
        result_query = session.query(flfiles_class).all()
        value_tmp = ""
        for file_ in result_query:
            value_tmp = util.sha1(file_.sha if not value_tmp else (value_tmp + file_.sha))

        value = value_tmp if value_tmp else value

        # session_dbaux = qsa.session("dbaux")
        session_dbaux = qsa.session()
        data_query = session_dbaux.query(flserial_class)

        if data_query.count():
            data_query.update({flserial_class.sha: value})
        else:
            obj_flserial = flserial_class()
            obj_flserial.sha = value
            session_dbaux.add(obj_flserial)

        return True
